import sys,copy
sys.path.append('..')

try:
    from MJOLNIRGui.src.main.python._tools import ProgressBarDecoratorArguments, loadUI
    import MJOLNIRGui.src.main.python._tools as _GUItools
    from MJOLNIRGui.src.main.python.DataModels import MaskModel
    from MJOLNIRGui.src.main.python.Views import BraggListManager
    #from MJOLNIRGui.src.main.python.MJOLNIR_Data import GuiMask
    
except ImportError:
    from DataModels import MaskModel
    #from MJOLNIR_Data import GuiMask
    from _tools import ProgressBarDecoratorArguments, loadUI
    import _tools as _GUItools
    from Views import BraggListManager
from MJOLNIR.Data import Mask
from PyQt5 import QtWidgets, uic, QtCore, QtGui
import numpy as np
import os,warnings
import inspect

class FilterProxyModel(QtCore.QIdentityProxyModel):
    def __init__(self,source=None, *args,**kwargs):
        super(FilterProxyModel,self).__init__(*args,**kwargs)
        if not source is None:
            self.setSourceModel(source)

    def flags(self,index):
        ds = self.sourceModel().data(index,role=QtCore.Qt.ItemDataRole)
        if len(ds._convertedFiles)>0:
            return  QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsEditable
        else:
            return QtCore.Qt.NoItemFlags


def Mask_DoubleClick_Selection_function(self,index,*args,**kwargs):
    self.Mask_listView.edit(index)

def setupMask(self):
    
    self.MaskModel = MaskModel(Mask_listView=self.Mask_listView,guiWindow=self.parent)
    self.Mask_listView.setModel(self.MaskModel)
    self.MaskModel.dataChanged.connect(self.MaskModel.layoutChanged.emit)

    self.MaskSelectionModel = self.Mask_listView.selectionModel()
    self.MaskSelectionModel.selectionChanged.connect(self.selectedMaskChanged)
    
    self.Mask_listView.doubleClicked.connect(self.Mask_DoubleClick_Selection_function)


def selectedMaskChanged(self,*args,**kwargs):
    self.updateMaskLabels()


def updateMaskLabels(self):
    masks = self.MaskModel.rowCount()
    if masks == 0:
        self.Mask_delete_button_nonBlock.setEnabled(False)
    else:
        self.Mask_delete_button_nonBlock.setEnabled(True)
    selected = self.Mask_listView.selectedIndexes()
    if len(np.asarray(selected))==0 or masks==0:
        self.Mask_duplicate_button_nonBlock.setEnabled(False)
        self.Mask_type_label.setText('')
    else:
        self.Mask_duplicate_button_nonBlock.setEnabled(True)
        if len(np.asarray(selected))>1:
            self.Mask_type_label.setText('Multiple')
        else:
            mask = self.MaskModel.item(np.asarray(selected)[0])
            maskType = type(mask).__name__
            self.Mask_type_label.setText(maskType)


def parse(string,masks):
    for _ObscureName in masks:
        locals()[_ObscureName.name]=_ObscureName
    return eval(string)


def MaskOnChange(self):
    self.Mask_parse_equation_text_edit_function()
    self.updateMaskLabels()


def Mask_parse_equation_text_edit_function(self):
    AppC = self.AppContext
    string = self.Mask_equation_line_edit.text()
    string = string.replace('\n','')
    string = string.strip()
    if len(string)==0:
        if hasattr(self.MaskModel,'combinedMask'):
            self.MaskModel.combinedMask = None
        img = QtGui.QPixmap(AppC.get_resource('Icons/Own/grey-led-on.png'))
    else:
        try:    
            combinedMask = parse(string,masks=self.MaskModel.masks)
        except (NameError,AttributeError,SyntaxError,TypeError):
            img = QtGui.QPixmap(AppC.get_resource('Icons/Own/red-led-on.png'))
            self.MaskModel.combinedMask = None
        else:
            if not isinstance(combinedMask,(Mask.MaskingObject)):
                img = QtGui.QPixmap(AppC.get_resource('Icons/Own/red-led-on.png'))
                self.MaskModel.combinedMask = None
            else:
                img = QtGui.QPixmap(AppC.get_resource('Icons/Own/green-led-on.png'))
                self.MaskModel.combinedMask = combinedMask
    img = img.scaled(QtCore.QSize(16, 16))
    self.Mask_status_label.setPixmap(img)
    self.maskingManager.performButtonChecks()

def Mask_add_button_function(self,oldMask=None):
    mask = self.maskingDialog(oldMask = oldMask)

    if mask is None:
        return
    else:
        self.MaskModel.append(mask)


def Mask_delete_button_function(self):
    self.MaskModel.delete(self.Mask_listView.selectedIndexes())
    self.updateMaskLabels()
    self.MaskModel.layoutChanged.emit()
    self.updateMaskLabels()
    

def Mask_duplicate_button_nonBlock_function(self):
    self.MaskModel.layoutChanged.emit()
    items = self.MaskModel.item(self.Mask_listView.selectedIndexes())
    if items is None:
        return
    items = np.asarray(items)
    for item in items:
        newItem = copy.deepcopy(item)
        self.MaskModel.append(newItem)
    
    self.updateMaskLabels()
    self.MaskModel.layoutChanged.emit()


### Define masks and their functions

maskTypes = {
    'Index Mask': 'loadIndexMask',
    'Line Mask': 'loadLineMask',
    'Rectangular Mask': 'loadRectangleMask',
    'Circular Mask': 'loadCircleMask',
    'Currat Axe Mask': 'loadCurratAxeMask',
    '3D Box Mask': 'loadBoxMask',
}

translate = {
    'indexMask': 'Index Mask',
    'lineMask': 'Line Mask',
    'rectangleMask': 'Rectangular Mask',
    'circleMask': 'Circular Mask',
    'CurratAxeMask': 'Currat Axe Mask',
    'boxMask': '3D Box Mask'
}

translateReverse = {}
for key,value in translate.items():
    translateReverse[value] = key


class maskingDialog(QtWidgets.QWidget):
    coordinates = ['h','k','l','energy','qx','qy','A3','A4','Ei']

    def __init__(self,  maskingManager, oldMask = None, *args, **kwargs):
        super(maskingDialog, self).__init__(*args, **kwargs)
        self.maskingManager = maskingManager
        self.MaskModel = self.maskingManager.maskingMainWindow.MaskModel
        self.oldMask = oldMask
        self.setWindowTitle("Masking")
        
        self.verticalLayout = QtWidgets.QVBoxLayout(self)
        self.verticalLayout.setObjectName("verticalLayout")
        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setObjectName("mainLayout")
        self.titleLayout = QtWidgets.QHBoxLayout()
        self.titleLayout.setObjectName("titleLayout")
        self.maskTypeComboBox = QtWidgets.QComboBox(self)
        self.maskTypeComboBox.setObjectName("maskTypeComboBox")
        self.maskTypeComboBox.addItem("")
        self.maskTypeComboBox.addItem("")
        self.titleLayout.addWidget(self.maskTypeComboBox)
        self.maskNameEdit = QtWidgets.QLineEdit(self)
        self.maskNameEdit.setObjectName("maskNameEdit")
        if not oldMask is None:
            self.maskNameEdit.setText(oldMask.name)
        self.titleLayout.addWidget(self.maskNameEdit)
        self.mainLayout.addLayout(self.titleLayout)
        self.inputLayoutOverview = QtWidgets.QHBoxLayout()
        self.inputLayoutOverview.setObjectName("inputLayoutOverview")

        self.cornerLayout = QtWidgets.QGridLayout()
        self.cornerLayout.setObjectName("cornerLayout")

        # self.coordinateLabel = QtWidgets.QLabel(self)
        # self.coordinateLabel.setObjectName("coordinateLabel")
        # self.inputLayoutOverview.addWidget(self.coordinateLabel)
        
        self.mainLayout.addLayout(self.inputLayoutOverview)
        self.mainLayout.addLayout(self.cornerLayout)
        self.verticalLayout.addLayout(self.mainLayout)

        self.initiateComboBox(mask=oldMask)
        if not oldMask is None:
            self.loadMaskType(translate[type(oldMask).__name__],oldMask)
        else:
            self.maskTypeComboBox.setCurrentIndex(0)
        
        QBtn = QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
        
        
        self.buttonBox = QtWidgets.QDialogButtonBox(QBtn)
        
        self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).setShortcut("Ctrl+Return")
        self.buttonBox.button(QtWidgets.QDialogButtonBox.Cancel).setShortcut("Esc")
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        

        
        self.verticalLayout.addWidget(self.buttonBox)
        self.setLayout(self.verticalLayout)

        self._changingBoxes = False
        

    def maskSelectionChanged(self,text):
        if text == '':
            return
        
        self.clearInputLayout() # Clear layout
        self.clearCornerLayout()
        self.loadMaskType(text)

    def loadMaskType(self,text,mask=None):
        if text in ['']:#self.simpleMaskParameters.keys():
            currentInfoParameters = self.simpleMaskParameters[text]
            dimensionality = currentInfoParameters[0][2]
            
            self.coordinateLabel = QtWidgets.QLabel('Coordinates: ')
            self.inputLayoutOverview.addWidget(self.coordinateLabel)
            self.coordinateComboBoxes = []
            for i in range(dimensionality):
                box = QtWidgets.QComboBox(self)
                box.setObjectName("coordinateBox"+str(i))
                self.inputLayoutOverview.addWidget(box)
                for coord in self.coordinates:
                    box.addItem(coord)
                
                self.coordinateComboBoxes.append(box)
            self.labels = [item[0] for item in currentInfoParameters]
            
            self.inputFields = []
            for i,label in enumerate(self.labels):
                localLabel = QtWidgets.QLabel(label)
                self.cornerLayout.addWidget(localLabel,i,0)
                for j,Type in enumerate(currentInfoParameters[i][1]):
                    if Type is float:
                        inputField = QtWidgets.QDoubleSpinBox()
                    elif Type is int:
                        inputField = QtWidgets.QSpinBox()
                    elif Type is bool:
                        inputField = QtWidgets.QCheckBox()
                    else:
                        raise AttributeError()
                    self.cornerLayout.addWidget(inputField,i,j+1)
                    self.inputFields.append(inputField)

        elif text in maskTypes:
            getattr(self,maskTypes[text])(mask=mask)
            
    def loadIndexMask(self,mask):
        labels = []
        self.widgets = []
        labels.append(QtWidgets.QLabel('Index Start'))
        self.widgets.append(QtWidgets.QSpinBox())
        
        labels.append(QtWidgets.QLabel('Index End'))
        self.widgets.append(QtWidgets.QSpinBox())
        
        labels.append(QtWidgets.QLabel('Axis'))
        axisComboBox=QtWidgets.QComboBox()
        axisComboBox.addItems(list(map(str,range(3))))
        axisComboBox.setCurrentIndex(0)
        self.widgets.append(axisComboBox)

        labels.append(QtWidgets.QLabel('Mask Inside'))
        self.widgets.append(QtWidgets.QCheckBox())

        params = ['start','end','axis','maskInside']
        if not mask is None:
            values = [getattr(mask,p) for p in params]
        else:
            values = [None for p in params]

        for i,(label,widget,value) in enumerate(zip(labels,self.widgets,values)):
            self.cornerLayout.addWidget(label,i,0)
            self.cornerLayout.addWidget(widget,i,1)
            if not value is None:
                if isinstance(widget,(QtWidgets.QSpinBox)):
                    widget.setValue(value)
                elif isinstance(widget,(QtWidgets.QComboBox)):
                    idx = widget.findText(str(value))
                    widget.setCurrentIndex(idx)
                elif isinstance(widget,(QtWidgets.QCheckBox)):
                    widget.setChecked(value)


        def acceptFunction(self):
            inputFields = {}
            inputFields['start'] = self.widgets[0].value()
            inputFields['end'] = self.widgets[1].value()
            inputFields['axis'] = int(self.widgets[2].currentText())
            inputFields['maskInside'] = self.widgets[3].isChecked()
            return inputFields

        self.acceptFunction = lambda: acceptFunction(self)

    def loadLineMask(self,mask):
        labels = []
        self.widgets = []

        labels.append(QtWidgets.QLabel('Coordinate'))
        axisComboBox=QtWidgets.QComboBox()
        axisComboBox.addItems(self.coordinates)
        axisComboBox.setCurrentIndex(0)
        self.widgets.append(axisComboBox)

        labels.append(QtWidgets.QLabel('Start'))
        DSBox = QtWidgets.QDoubleSpinBox()
        DSBox.setMaximum(1e6)
        DSBox.setMinimum(-1e6)
        self.widgets.append(DSBox)
        
        labels.append(QtWidgets.QLabel('End'))
        DSBox = QtWidgets.QDoubleSpinBox()
        DSBox.setMaximum(1e6)
        DSBox.setMinimum(-1e6)
        self.widgets.append(DSBox)
        
        

        labels.append(QtWidgets.QLabel('Mask Inside'))
        self.widgets.append(QtWidgets.QCheckBox())

        params = ['coordinates','start','end','maskInside']
        if not mask is None:
            values = [getattr(mask,p) for p in params]
        else:
            values = [None for p in params]

        for i,(label,widget,value) in enumerate(zip(labels,self.widgets,values)):
            self.cornerLayout.addWidget(label,i,0)
            self.cornerLayout.addWidget(widget,i,1)
            if not value is None:
                if isinstance(widget,(QtWidgets.QSpinBox,QtWidgets.QDoubleSpinBox)):
                    widget.setValue(value)
                elif isinstance(widget,(QtWidgets.QComboBox)):
                    idx = widget.findText(str(value[0]))
                    widget.setCurrentIndex(idx)
                elif isinstance(widget,(QtWidgets.QCheckBox)):
                    widget.setChecked(value)


        def acceptFunction(self):
            inputFields = {}
            inputFields['start'] = self.widgets[1].value()
            inputFields['end'] = self.widgets[2].value()
            inputFields['coordinates'] = [self.widgets[0].currentText()]
            inputFields['maskInside'] = self.widgets[3].isChecked()
            return inputFields

        self.acceptFunction = lambda: acceptFunction(self)

    def loadRectangleMask(self,mask):
        labels = []
        self.widgets = []
        
        
        labels.append(QtWidgets.QLabel('Coordinates'))
        axisComboBox=QtWidgets.QComboBox()
        axisComboBox.addItems(self.coordinates)
        axisComboBox.setCurrentIndex(0)
        self.widgets.append(axisComboBox)
        axisComboBox=QtWidgets.QComboBox()
        axisComboBox.addItems(self.coordinates)
        axisComboBox.setCurrentIndex(1)
        self.widgets.append(axisComboBox)

        labels.append(QtWidgets.QLabel('Corner 1'))
        DSBox = QtWidgets.QDoubleSpinBox()
        DSBox.setMaximum(1e6)
        DSBox.setMinimum(-1e6)
        self.widgets.append(DSBox)
        DSBox = QtWidgets.QDoubleSpinBox()
        DSBox.setMaximum(1e6)
        DSBox.setMinimum(-1e6)
        self.widgets.append(DSBox)
        
        labels.append(QtWidgets.QLabel('Corner 2'))
        DSBox = QtWidgets.QDoubleSpinBox()
        DSBox.setMaximum(1e6)
        DSBox.setMinimum(-1e6)
        self.widgets.append(DSBox)
        DSBox = QtWidgets.QDoubleSpinBox()
        DSBox.setMaximum(1e6)
        DSBox.setMinimum(-1e6)
        self.widgets.append(DSBox)

        self.corner3CheckBox = QtWidgets.QCheckBox()
        self.corner3CheckBox.setText('Corner 3')
        
        def onToggle(value,self):
            
            for spinBox in self.widgets[6:8]:
                if value:
                    spinBox.setDisabled(False)
                else:
                    spinBox.setDisabled(True)
        self.corner3CheckBox.toggled.connect(lambda value: onToggle(value,self))
        
        
        labels.append(self.corner3CheckBox)
        DSBox = QtWidgets.QDoubleSpinBox()
        DSBox.setMaximum(1e6)
        DSBox.setMinimum(-1e6)
        self.widgets.append(DSBox)
        DSBox = QtWidgets.QDoubleSpinBox()
        DSBox.setMaximum(1e6)
        DSBox.setMinimum(-1e6)
        self.widgets.append(DSBox)
        
        self.corner3CheckBox.setChecked(True)
        
        labels.append(QtWidgets.QLabel('Mask Inside'))
        self.widgets.append(QtWidgets.QCheckBox())

        params = ['coordinates','corner1','corner2','corner3','maskInside']
        if not mask is None:
            values = [getattr(mask,p) for p in params]
        else:
            values = [None for p in params]

        for i,(label,value) in enumerate(zip(labels,values)):
            self.cornerLayout.addWidget(label,i,0)
            w1 = self.widgets[2*i]
            self.cornerLayout.addWidget(w1,i,1)
            if len(self.widgets)!=i*2+1:
                w2 = self.widgets[2*i+1]
                self.cornerLayout.addWidget(w2,i,1+1)
            else:
                w2 = None

            if not value is None:
                if isinstance(w1,(QtWidgets.QSpinBox,QtWidgets.QDoubleSpinBox)):
                    w1.setValue(value[0])
                    w2.setValue(value[1])
                elif isinstance(w1,(QtWidgets.QComboBox)):
                    w1.setCurrentIndex(w1.findText(str(value[0])))
                    w1.setCurrentIndex(w2.findText(str(value[1])))
                elif isinstance(w1,(QtWidgets.QCheckBox)):
                    w1.setChecked(value)


        def acceptFunction(self):
            inputFields = {}
            inputFields['coordinates'] = [self.widgets[0].currentText(),self.widgets[1].currentText()]
            inputFields['corner1'] = [self.widgets[2].value(),self.widgets[3].value()]
            inputFields['corner2'] = [self.widgets[4].value(),self.widgets[5].value()]
            if self.corner3CheckBox.isChecked():
                inputFields['corner3'] = [self.widgets[6].value(),self.widgets[7].value()]
            inputFields['maskInside'] = self.widgets[8].isChecked()
            return inputFields

        self.acceptFunction = lambda: acceptFunction(self)
    
    def loadBoxMask(self,mask):
        labels = []
        self.widgets = []
        dimensionality = 3
        def onToggle(value,widgets):
            
            for spinBox in widgets:
                if value:
                    spinBox.setDisabled(False)
                else:
                    spinBox.setDisabled(True)
        
        labels.append(QtWidgets.QLabel('Coordinates'))
        for i in range(dimensionality):
            axisComboBox=QtWidgets.QComboBox()
            axisComboBox.addItems(self.coordinates)
            axisComboBox.setCurrentIndex(i)
            self.widgets.append(axisComboBox)

        labels.append(QtWidgets.QLabel('Corner 1'))
        for _ in range(dimensionality):
            DSBox = QtWidgets.QDoubleSpinBox()
            DSBox.setMaximum(1e6)
            DSBox.setMinimum(-1e6)
            self.widgets.append(DSBox)

        labels.append(QtWidgets.QLabel('Corner 2'))
        for _ in range(dimensionality):
            DSBox = QtWidgets.QDoubleSpinBox()
            DSBox.setMaximum(1e6)
            DSBox.setMinimum(-1e6)
            self.widgets.append(DSBox)
        
        labels.append(QtWidgets.QLabel('Corner 3'))
        for _ in range(dimensionality):
            DSBox = QtWidgets.QDoubleSpinBox()
            DSBox.setMaximum(1e6)
            DSBox.setMinimum(-1e6)
            self.widgets.append(DSBox)

        
        self.corner4CheckBox = QtWidgets.QCheckBox()
        self.corner4CheckBox.setText('Corner 4')
        labels.append(self.corner4CheckBox)

        for _ in range(dimensionality):
            
            DSBox = QtWidgets.QDoubleSpinBox()
            DSBox.setMaximum(1e6)
            DSBox.setMinimum(-1e6)
            self.widgets.append(DSBox)
        widgets = len(self.widgets)
        self.corner4CheckBox.toggled.connect(lambda value: onToggle(value,self.widgets[widgets-dimensionality:widgets]))
        
        self.corner4CheckBox.setChecked(False)
        self.corner4CheckBox.setChecked(True)
        
        labels.append(QtWidgets.QLabel('Mask Inside'))
        self.widgets.append(QtWidgets.QCheckBox())

        params = ['coordinates','corner1','corner2','corner3','corner4','maskInside']
        sizes = [dimensionality,dimensionality,dimensionality,dimensionality,dimensionality,1]
        if not mask is None:
            values = [getattr(mask,p) for p in params]
        else:
            values = [[None]*s for s in sizes]

        for i,(label,value) in enumerate(zip(labels,values)):
            self.cornerLayout.addWidget(label,i,0)

            if isinstance(value,(bool)):
                value = [value]
            elif isinstance(value,(np.ndarray)):
                if np.any(np.isnan(value)):
                    value = [None]*dimensionality
            for j,val in enumerate(value):
                w = self.widgets[i*dimensionality+j]
                self.cornerLayout.addWidget(w,i,j+1)
                if not val is None:
                    if isinstance(w,(QtWidgets.QSpinBox,QtWidgets.QDoubleSpinBox)):
                        w.setValue(val)
                    elif isinstance(w,(QtWidgets.QComboBox)):
                        w.setCurrentIndex(w.findText(str(val)))
                    elif isinstance(w,(QtWidgets.QCheckBox)):
                        w.setChecked(val)

        def acceptFunction(self):
            inputFields = {}
            inputFields['coordinates'] = [x.currentText() for x in self.widgets[:dimensionality]]
            inputFields['corner1'] = [x.value() for x in self.widgets[1*dimensionality:2*dimensionality]]
            inputFields['corner2'] = [x.value() for x in self.widgets[2*dimensionality:3*dimensionality]]
            inputFields['corner3'] = [x.value() for x in self.widgets[3*dimensionality:4*dimensionality]]
            if self.corner4CheckBox.isChecked():
                inputFields['corner4'] = [x.value() for x in self.widgets[4*dimensionality:5*dimensionality]]
            inputFields['maskInside'] = self.widgets[-1].isChecked()
            return inputFields

        self.acceptFunction = lambda: acceptFunction(self)


    def loadCircleMask(self,mask):
        labels = []
        self.widgets = []
        dimensionality = 2
        
        labels.append(QtWidgets.QLabel('Coordinates'))
        for i in range(dimensionality):
            axisComboBox=QtWidgets.QComboBox()
            axisComboBox.addItems(self.coordinates)
            axisComboBox.setCurrentIndex(i)
            self.widgets.append(axisComboBox)

        labels.append(QtWidgets.QLabel('Center'))
        for _ in range(dimensionality):
            DSBox = QtWidgets.QDoubleSpinBox()
            DSBox.setMaximum(1e6)
            DSBox.setMinimum(-1e6)
            self.widgets.append(DSBox)

        labels.append(QtWidgets.QLabel('Radius Point'))
        for _ in range(dimensionality):
            DSBox = QtWidgets.QDoubleSpinBox()
            DSBox.setMaximum(1e6)
            DSBox.setMinimum(-1e6)
            self.widgets.append(DSBox)

        labels.append(QtWidgets.QLabel('Mask Inside'))
        self.widgets.append(QtWidgets.QCheckBox())
        

        params = ['coordinates','center','radiusPoint','maskInside']
        sizes = [dimensionality,dimensionality,dimensionality,1]
        if not mask is None:
            values = [getattr(mask,p).flatten() if hasattr(getattr(mask,p),'flatten') else getattr(mask,p) for p in params]
        else:
            values = [[None]*s for s in sizes]

        for i,(label,value) in enumerate(zip(labels,values)):
            self.cornerLayout.addWidget(label,i,0)

            if isinstance(value,(bool)):
                value = [value]
            elif isinstance(value,(np.ndarray)):
                if np.any(np.isnan(value)):
                    value = [None]*dimensionality
            if i == 3:
                if not hasattr(value,'__len__'): # then it is not [None]
                    value = [value]
            for j,val in enumerate(value):
                if i < 3:
                    index =i*dimensionality+j
                else:
                    index = 2*dimensionality+i-1

                w = self.widgets[index]
                self.cornerLayout.addWidget(w,i,j+1)
                if not val is None:
                    if isinstance(w,(QtWidgets.QSpinBox,QtWidgets.QDoubleSpinBox)):
                        w.setValue(val)
                    elif isinstance(w,(QtWidgets.QComboBox)):
                        w.setCurrentIndex(w.findText(str(val)))
                    elif isinstance(w,(QtWidgets.QCheckBox)):
                        w.setChecked(val)

        def acceptFunction(self):
            inputFields = {}
            inputFields['coordinates'] = [x.currentText() for x in self.widgets[:dimensionality]]
            inputFields['center'] = [x.value() for x in self.widgets[1*dimensionality:2*dimensionality]]
            inputFields['radiusPoint'] = [x.value() for x in self.widgets[2*dimensionality:3*dimensionality]]
            #inputFields['radius'] = self.widgets[3*dimensionality].value()
            inputFields['maskInside'] = self.widgets[-1].isChecked()
            return inputFields

        self.acceptFunction = lambda: acceptFunction(self)

    def loadCurratAxeMask(self,mask):
        if not mask is None:
            braggList = mask.braggPeaks
        else:
            braggList = None
        

        self.BraggListManager = BraggListManager.BraggListManager(parent=self,BraggList=braggList)
        self.cornerLayout.addWidget(self.BraggListManager,0,0)
        ###
        self.verticalLayoutWidget = QtWidgets.QWidget(self)
        self.verticalLayout = QtWidgets.QVBoxLayout(self.verticalLayoutWidget)
        
        
        self.verticalLayout2 = QtWidgets.QVBoxLayout()
        
        self.verticalLayout2.addWidget(QtWidgets.QLabel('Mask Size'))

        self.buttonGroup = QtWidgets.QButtonGroup(self.verticalLayoutWidget)
        self.buttonGroup.setObjectName(u"buttonGroup")
        self.QxQy_radioButton = QtWidgets.QRadioButton(self.verticalLayoutWidget)
        self.buttonGroup.addButton(self.QxQy_radioButton)
        
        self.verticalLayout2.addWidget(self.QxQy_radioButton)

        self.RLU_radioButton = QtWidgets.QRadioButton(self.verticalLayoutWidget)
        self.buttonGroup.addButton(self.RLU_radioButton)
        self.RLU_radioButton.setChecked(True)

        self.verticalLayout2.addWidget(self.RLU_radioButton)

        self.verticalLayout.addLayout(self.verticalLayout2)

        self.QxQy_radioButton.setText("Qx Qy")
        self.RLU_radioButton.setText("RLU")

        self.gridLayout = QtWidgets.QGridLayout()
        self.BraggList_dH_spinBox = QtWidgets.QDoubleSpinBox(self.verticalLayoutWidget)
        self.BraggList_dH_spinBox.setObjectName(u"BraggList_H_spinBox")
        self.BraggList_dH_spinBox.setDecimals(3)
        self.BraggList_dH_spinBox.setMinimum(0.001)
        self.BraggList_dH_spinBox.setMaximum(99.998999999999995)
        self.BraggList_dH_spinBox.setSingleStep(0.010000000000000)
        self.BraggList_dH_spinBox.setValue(0.100000000000000)

        self.gridLayout.addWidget(self.BraggList_dH_spinBox, 1, 0, 1, 1)

        self.BraggList_dK_spinBox = QtWidgets.QDoubleSpinBox(self.verticalLayoutWidget)
        self.BraggList_dK_spinBox.setObjectName(u"BraggList_K_spinBox")
        self.BraggList_dK_spinBox.setDecimals(3)
        self.BraggList_dK_spinBox.setMinimum(0.001)
        self.BraggList_dK_spinBox.setMaximum(99.998999999999995)
        self.BraggList_dK_spinBox.setSingleStep(0.010000000000000)
        self.BraggList_dK_spinBox.setValue(0.100000000000000)

        self.gridLayout.addWidget(self.BraggList_dK_spinBox, 1, 1, 1, 1)

        self.BraggList_dL_spinBox = QtWidgets.QDoubleSpinBox(self.verticalLayoutWidget)
        self.BraggList_dL_spinBox.setObjectName(u"BraggList_L_spinBox")
        self.BraggList_dL_spinBox.setDecimals(3)
        self.BraggList_dL_spinBox.setMinimum(0.001)
        self.BraggList_dL_spinBox.setMaximum(99.998999999999995)
        self.BraggList_dL_spinBox.setSingleStep(0.010000000000000)
        self.BraggList_dL_spinBox.setValue(0.100000000000000)

        self.gridLayout.addWidget(self.BraggList_dL_spinBox, 1, 2, 1, 1)

        self.BraggListQH_label = QtWidgets.QLabel(self.verticalLayoutWidget)
        self.BraggListQH_label.setText('dH')
        self.BraggListQH_label.setAlignment(QtCore.Qt.AlignCenter)
        self.gridLayout.addWidget(self.BraggListQH_label, 0, 0, 1, 1)

        self.BraggListQK_label = QtWidgets.QLabel(self.verticalLayoutWidget)
        self.BraggListQK_label.setText('dK')
        self.BraggListQK_label.setAlignment(QtCore.Qt.AlignCenter)
        self.gridLayout.addWidget(self.BraggListQK_label, 0, 1, 1, 1)

        self.BraggListQL_label = QtWidgets.QLabel(self.verticalLayoutWidget)
        self.BraggListQL_label.setText('dL')
        self.BraggListQL_label.setAlignment(QtCore.Qt.AlignCenter)

        self.gridLayout.addWidget(self.BraggListQL_label, 0, 2, 1, 1)


        self.verticalLayout.addLayout(self.gridLayout)

        self.horizontalLayout = QtWidgets.QVBoxLayout()
        
        self.horizontalLayout.addWidget(QtWidgets.QLabel('Mask Inside'))

        self.MaskInside_checkBox = QtWidgets.QCheckBox()
        if not mask is None:
            self.MaskInside_checkBox.setChecked(mask.maskInside)
        self.horizontalLayout.addWidget(self.MaskInside_checkBox)

        self.cornerLayout.addWidget(self.verticalLayoutWidget,1,0)

        def toggle(self,value): # To be added to QxQy
            if value:
                self.BraggList_dL_spinBox.setDisabled(True)
                self.BraggListQH_label.setText('dqx')
                self.BraggListQK_label.setText('dqx')
                self.BraggListQL_label.setText('')
            else:
                self.BraggList_dL_spinBox.setDisabled(False)
                self.BraggListQH_label.setText('dH')
                self.BraggListQK_label.setText('dK')
                self.BraggListQL_label.setText('dL')
        
        self.QxQy_radioButton.toggled.connect(lambda value: toggle(self,value))
        if not mask is None:
            dqx,dqy, dH, dK, dL = mask.dqx,mask.dqy, mask.dH, mask.dK, mask.dL
            if np.any([x is None for x in [dqx,dqy]]): # Then use dH, dK, dL
                self.BraggList_dH_spinBox.setValue(dH)
                self.BraggList_dK_spinBox.setValue(dK)
                self.BraggList_dL_spinBox.setValue(dL)
            else:
                self.QxQy_radioButton.setChecked(True)
                self.BraggList_dH_spinBox.setValue(dqx)
                self.BraggList_dK_spinBox.setValue(dqy)



        def acceptFunction(self):
            inputFields = {}
            inputFields['braggPeaks'] = self.BraggListManager.BraggListModel.data
            if self.QxQy_radioButton.isChecked():
                inputFields['dqx'] = self.BraggList_dH_spinBox.value()
                inputFields['dqx'] = self.BraggList_dK_spinBox.value()
            else:
                inputFields['dH'] = self.BraggList_dH_spinBox.value()
                inputFields['dK'] = self.BraggList_dK_spinBox.value()
                inputFields['dL'] = self.BraggList_dL_spinBox.value()
            inputFields['maskInside'] = self.MaskInside_checkBox.isChecked()
            return inputFields

        self.acceptFunction = lambda: acceptFunction(self)

        
    def coordinateSelectionChanged(self,id,text):
        if self._changingBoxes or text == '': # If already changing the boxes through updating of items
            return
        
        self._changingBoxes = True # Start changing the boxes
        totalCoordinates = self.coordinates.copy()
        currentCoordinates = [box.currentText() for box in self.coordinateComboBoxes]
        for box,prevCoord,ownCoord in zip(self.coordinateComboBoxes[1:],currentCoordinates,currentCoordinates[1:]):
            box.clear()
            totalCoordinates.remove(prevCoord)
            box.addItems(totalCoordinates)

        self._changingBoxes = False
            
        

    def clearInputLayout(self):
        for i in reversed(range(self.inputLayoutOverview.count())): 
            self.inputLayoutOverview.itemAt(i).widget().setParent(None)
    
    def clearCornerLayout(self):
        for i in reversed(range(self.cornerLayout.count())): 
            self.cornerLayout.itemAt(i).widget().setParent(None)

    def accept(self): # the accept button has been pressed
        # coordinates = [box.currentText() for box in self.coordinateComboBoxes]
        MaskType = getattr(Mask,translateReverse[self.maskTypeComboBox.currentText()])
        
        maskName = self.maskNameEdit.text()
        if len(maskName) == 0:
            maskName = None
        inputFields = self.acceptFunction()

        self.mask = MaskType(**inputFields,name=maskName)
        if not self.oldMask is None:
            idx = self.MaskModel.getNames().index(self.oldMask.name)
            index = self.MaskModel.index(idx)
            self.MaskModel.delete([index])
        self.MaskModel.append(self.mask)
        self.close()
        #return super(maskingDialog,self).accept()

    def reject(self):
        self.close()

    def initiateComboBox(self,mask=None):
        self.maskTypeComboBox.clear()  # Needed to clear empty input
        names = list(maskTypes.keys())#['indexMask','lineMask','rectangleMask','circleMask','boxMask','CurratAxeMask']#self.simpleMaskParameters.keys()
        for name in names:
            self.maskTypeComboBox.addItem(name)
        if not mask is None:
            self.maskTypeComboBox.setCurrentIndex(names.index(translate[type(mask).__name__]))
        else:
            self.maskTypeComboBox.setCurrentIndex(-1)

        self.maskTypeComboBox.currentTextChanged.connect(self.maskSelectionChanged)

class AnotherWindow(QtWidgets.QWidget):
    """
    Create class to handle second window
    """
    def __init__(self,widget,AppContext,parent,maskingManager):
        super().__init__()
        self.maskingManager = maskingManager
        self.AppContext = AppContext
        layout = QtWidgets.QVBoxLayout()
        
        self.setLayout(layout)
        layout.addWidget(widget)
        self.parent = parent

        self.dialog = None

    # Method to generate masking dialogue
    def maskingDialog(self,oldMask=None):
        self.dialog = maskingDialog(maskingManager=self.maskingManager,oldMask=oldMask)

        if self.dialog.show(): # Execute the dialog
            return None
        
    def closeEvent(self, event):
        if not self.parent.isClosing:
            self.setVisible(False)
            if not self.dialog is None:
                self.dialog.close()
            return
        else:
            if not self.dialog is None:
                self.dialog.close()
            
        
@ProgressBarDecoratorArguments(runningText='Applying Masking',completedText='Masking Applied')
def ApplyMaskToDataSet(self):
    self.resetProgressBar()
    self.setProgressBarMaximum(1)
    self.setProgressBarLabelText('Applying Masking')
    localIdx = self.maskingManager.Mask_data_set_combo_box.currentIndex()
    idx = self.DataSetModel.index(localIdx,0) # Create correct index type for DataSetModel
    ds = self.DataSetModel.data(idx,QtCore.Qt.ItemDataRole) # Get the DataSet in question
    self.setProgressBarLabelText('Applying Masking to "{}"'.format(ds.name))

    
    mask = self.maskingManager.getMasks()
    matrix = mask(ds)

    
    self.addProgressBarValue(0.75)
    
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        ds.mask = matrix
        if len(w)>0:
                self.writeToStatus(str(w[0].message))

    ds._maskingObject = mask
    
    self.maskingManager.closeEvent(event=None)    
    return True


MaskManagerBase, MaskManagerForm = loadUI("Mask.ui")

class MaskManager(MaskManagerBase, MaskManagerForm):
    def __init__(self, parent=None):
        super(MaskManager, self).__init__(parent)
        self.parent = parent
        self.setupUi(self)
        
        self.maskingMainWindow = AnotherWindow(widget=self,AppContext=self.parent.AppContext,parent=self.parent,maskingManager=self)
        self.guiWindow = self.parent
        self.maskingMainWindow.setVisible(False)
        self.MaskIsCorrect = False
        self.initMaskManager()
        self.setup()
        self.setupContextMenu()

    def setupContextMenu(self):
        def contextMenu(view,event,self):
            # Generate a context menu that opens on right click
            position = event.globalPos()
            idx = view.selectedIndexes()
            
            if len(idx)!=0:
                mask = view.model().item(idx[0])
                if event.type() == QtCore.QEvent.ContextMenu:
                    menu = QtWidgets.QMenu()
                    edit = QtWidgets.QAction('Edit')
                    edit.setToolTip('Edit Mask') 
                    edit.setStatusTip(edit.toolTip())
                    edit.triggered.connect(lambda: Mask_add_button_function(self.maskingMainWindow,oldMask=mask))
                    edit.setIcon(QtGui.QIcon(self.parent.AppContext.get_resource('Icons/Own/mask-edit.png')))
                    menu.addAction(edit)

                    duplicate = QtWidgets.QAction('Duplicate')
                    duplicate.setToolTip('Duplicate Mask') 
                    duplicate.setStatusTip(duplicate.toolTip())
                    duplicate.triggered.connect(lambda: Mask_duplicate_button_nonBlock_function(self.maskingMainWindow))
                    duplicate.setIcon(QtGui.QIcon(self.parent.AppContext.get_resource('Icons/Own/mask-duplicate.png')))
                    menu.addAction(duplicate)

                    delete = QtWidgets.QAction('Delete')
                    delete.setToolTip('Delete Mask') 
                    delete.setStatusTip(delete.toolTip())
                    delete.triggered.connect(lambda: Mask_delete_button_function(self.maskingMainWindow))
                    delete.setIcon(QtGui.QIcon(self.parent.AppContext.get_resource('Icons/Own/cross-button.png')))
                    menu.addAction(delete)


                    return menu.exec_(position)

        self.maskingMainWindow.MaskModel.Mask_listView.contextMenuEvent = lambda event: contextMenu(self.maskingMainWindow.MaskModel.Mask_listView,event,self)


    def setWindowVisible(self):
        self.maskingMainWindow.setVisible(True)

    def initMaskManager(self):
        self.maskingMainWindow.Mask_add_button_function = lambda: Mask_add_button_function(self.maskingMainWindow)
        self.maskingMainWindow.Mask_duplicate_button_nonBlock_function = lambda: Mask_duplicate_button_nonBlock_function(self.maskingMainWindow)
        self.maskingMainWindow.Mask_parse_equation_text_edit_function = lambda: Mask_parse_equation_text_edit_function(self.maskingMainWindow)
        self.maskingMainWindow.setupMask = lambda: setupMask(self.maskingMainWindow)

        self.maskingMainWindow.Mask_DoubleClick_Selection_function = lambda index:Mask_DoubleClick_Selection_function(self.maskingMainWindow,index)
        self.maskingMainWindow.Mask_delete_button_function = lambda:Mask_delete_button_function(self.maskingMainWindow)
        self.maskingMainWindow.selectedMaskChanged = lambda : selectedMaskChanged(self.maskingMainWindow)
        self.maskingMainWindow.updateMaskLabels = lambda:updateMaskLabels(self.maskingMainWindow)
        self.maskingMainWindow.MaskOnChange = lambda:MaskOnChange(self.maskingMainWindow)

        self.guiWindow.ApplyMaskToDataSet = lambda: ApplyMaskToDataSet(self.guiWindow)

        for key,value in self.__dict__.items():
                if 'Mask' in key:
                    self.maskingMainWindow.__dict__[key] = value

        grey = '<img src="{}" width=16 height=16>'.format(self.parent.AppContext.get_resource('Icons/Own/grey-led-on.png'))
        red = '<img src="{}" width=16 height=16>'.format(self.parent.AppContext.get_resource('Icons/Own/red-led-on.png'))
        green = '<img src="{}" width=16 height=16>'.format(self.parent.AppContext.get_resource('Icons/Own/green-led-on.png'))
        img = QtGui.QPixmap(self.parent.AppContext.get_resource('Icons/Own/grey-led-on.png'))
        img = img.scaled(QtCore.QSize(16, 16))
        self.maskingMainWindow.Mask_status_label.setPixmap(img)
        self.maskingMainWindow.Mask_status_label.setToolTip('<br>'.join([grey+' No mask applied',red+' Wrong masking',green+' Valid masking']))
        
        self.proxyModel = FilterProxyModel(self.parent.DataSetModel)

        self.Mask_data_set_combo_box.setModel(self.proxyModel)
        self.Mask_apply_to_dataset_button.clicked.connect(self.guiWindow.ApplyMaskToDataSet)
        self.Mask_load_current_button_nonBlock.clicked.connect(self.Mask_load_current_button_nonBlock_function)
        self.Mask_data_set_combo_box.currentIndexChanged.connect(self.performButtonChecks)
        self.parent.DataSetModel.layoutChanged.connect(self.DataSetModelLayoutChanged)

        self.performButtonChecks()


        
    def setup(self):
        self.maskingMainWindow.setupMask()
        self.maskingMainWindow.updateMaskLabels()
        self.maskingMainWindow.Mask_delete_button_nonBlock.clicked.connect(self.maskingMainWindow.Mask_delete_button_function)
        self.maskingMainWindow.Mask_delete_button_nonBlock.setShortcut("Del")

        self.maskingMainWindow.Mask_add_button_nonBlock.clicked.connect(self.maskingMainWindow.Mask_add_button_function)
        self.maskingMainWindow.Mask_add_button_nonBlock.setShortcut("Ctrl+N")

        self.maskingMainWindow.Mask_duplicate_button_nonBlock.clicked.connect(self.maskingMainWindow.Mask_duplicate_button_nonBlock_function)
        self.maskingMainWindow.Mask_equation_line_edit.textChanged.connect(self.maskingMainWindow.Mask_parse_equation_text_edit_function)
        self.maskingMainWindow.MaskModel.layoutChanged.connect(self.maskingMainWindow.MaskOnChange)
        self.maskingMainWindow.Mask_save_file_button.clicked.connect(self.save_mask_file)
        self.maskingMainWindow.Mask_save_file_button.setShortcut("Ctrl+S")

        self.maskingMainWindow.Mask_load_file_button.clicked.connect(self.load_mask_file)
        self.maskingMainWindow.Mask_load_file_button.setShortcut("Ctrl+O")

    def getMasks(self):
        if hasattr(self.maskingMainWindow,'MaskModel'):
            return self.maskingMainWindow.MaskModel.combinedMask
        else:
            return None
        
        
    def closeEvent(self, event):
        self.maskingMainWindow.close()


    def DataSetModelLayoutChanged(self):
        # Make clever check in order to not choose a non-valid ds
        if self.parent.DataSetModel.rowCount(None)==0:
            self.Mask_data_set_combo_box.setCurrentIndex(-1)
            return 
        
        currentIndex = self.Mask_data_set_combo_box.currentIndex()
        if currentIndex == -1:
            return
        currentFlags = self.Mask_data_set_combo_box.model().flags(self.Mask_data_set_combo_box.model().index(currentIndex,0))

        if not bool(currentFlags & QtCore.Qt.ItemIsEnabled):
            self.Mask_data_set_combo_box.setCurrentIndex(-1)
            
        #else:
        #    self.Mask_data_set_combo_box.setCurrentIndex(-1)

        self.performButtonChecks()
        
    

    def performButtonChecks(self):
        self.Mask_apply_to_dataset_button_check()
        self.Mask_load_current_button_nonBlock_check()
        self.Mask_save_file_button_check()

    def Mask_load_current_button_nonBlock_check(self):
        if self.parent.DataSetModel.rowCount(None)!=0:
            localIdx = self.Mask_data_set_combo_box.currentIndex()
            idx = self.parent.DataSetModel.index(localIdx,0)
            ds = self.parent.DataSetModel.data(idx,QtCore.Qt.ItemDataRole)
            if not ds._maskingObject is None:
                self.Mask_load_current_button_nonBlock.setDisabled(False)
            else:
                self.Mask_load_current_button_nonBlock.setDisabled(True)
        else:
            self.Mask_load_current_button_nonBlock.setDisabled(True)

    def Mask_apply_to_dataset_button_check(self):
        if self.parent.DataSetModel.rowCount(None)!=0 and self.getMasks() is not None:
            localIdx = self.Mask_data_set_combo_box.currentIndex()
            idx = self.parent.DataSetModel.index(localIdx,0)
            ds = self.parent.DataSetModel.data(idx,QtCore.Qt.ItemDataRole)

            if len(ds._convertedFiles)>0 and localIdx>-1: # it has converted files and is selected
                self.Mask_apply_to_dataset_button.setDisabled(False)
            else:
                self.Mask_apply_to_dataset_button.setDisabled(True)
        else:
            self.Mask_apply_to_dataset_button.setDisabled(True)

    def Mask_save_file_button_check(self):
        if self.getMasks() is not None:
            self.Mask_save_file_button.setDisabled(False)
        else:
            self.Mask_save_file_button.setDisabled(True)

    def Mask_load_current_button_nonBlock_function(self):
        localIdx = self.Mask_data_set_combo_box.currentIndex()

        idx = self.parent.DataSetModel.index(localIdx,0)
        ds = self.parent.DataSetModel.data(idx,QtCore.Qt.ItemDataRole)

        eq,masks = Mask.extract(ds._maskingObject)

        currentNames = self.maskingMainWindow.MaskModel.getNames()
        newNames = [m.name for m in masks]

        overlap = list(set.intersection(set(currentNames),set(newNames)))
        if len(overlap)>0:
            dialog = OverlapDialog(names=overlap)
            result = dialog.exec_()
            if result:
                ### Delete all masks with names in 
                names = list([m.name for m in self.maskingMainWindow.MaskModel.masks])

                idx =  [names.index(name) for name in overlap] # make sure ascending
                index = self.maskingMainWindow.MaskModel.index(-1,0)
                self.maskingMainWindow.MaskModel.Mask_listView.setCurrentIndex(index)
                self.maskingMainWindow.MaskModel.delete([self.maskingMainWindow.MaskModel.index(i,0) for i in idx])
            else:
                return False

        for m in masks:
            self.maskingMainWindow.MaskModel.append(m)

        self.maskingMainWindow.Mask_equation_line_edit.setText(eq)
        self.performButtonChecks()
        return True

    def save_mask_file(self):
        mask = self.getMasks()
        saveFile,_ = QtWidgets.QFileDialog.getSaveFileName(self, 'Save File')

        if saveFile is None or saveFile == '':
            return False

        if not saveFile.split('.')[-1] == 'mask':
            saveFile+='.mask'

        if os.path.exists(saveFile): # if file already exists user has been asked to delete it
            os.remove(saveFile)
        
        mask.save(saveFile)

    def load_mask_file(self):

        loadFile,_ = QtWidgets.QFileDialog.getOpenFileName(self,"Load Mask", None ,"Mask (*.mask);;All Files (*)")

        if loadFile is None or loadFile == '':
            return False

        mask = Mask.load(loadFile)

        eq,masks = Mask.extract(mask)

        currentNames = self.maskingMainWindow.MaskModel.getNames()
        newNames = [m.name for m in masks]

        overlap = list(set.intersection(set(currentNames),set(newNames)))
        if len(overlap)>0:
            dialog = OverlapDialog(names=overlap)
            result = dialog.exec_()
            if result:
                ### Delete all masks with names in 
                names = list([m.name for m in self.maskingMainWindow.MaskModel.masks])

                idx =  [names.index(name) for name in overlap] # make sure ascending
                index = self.maskingMainWindow.MaskModel.index(-1,0)
                self.maskingMainWindow.MaskModel.Mask_listView.setCurrentIndex(index)
                self.maskingMainWindow.MaskModel.delete([self.maskingMainWindow.MaskModel.index(i,0) for i in idx])
            else:
                return False

        for m in masks:
            self.maskingMainWindow.MaskModel.append(m)

        self.maskingMainWindow.Mask_equation_line_edit.setText(eq)
        self.performButtonChecks()
        return True

        


class OverlapDialog(QtWidgets.QDialog):
    def __init__(self,names,*args,**kwargs):
        super().__init__(*args,**kwargs)

        self.names = names
        self.setWindowTitle("Name{:} already in use!".format('s'*(len(names)>1)))

        QBtn = QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
        self.buttonBox = QtWidgets.QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)


        self.layout = QtWidgets.QVBoxLayout()
        message = QtWidgets.QLabel("Loaded mask contains name{0} already in use. \n Overwrite with the loaded mask{0}?".format('s'*(len(names)>1)))
        self.layout.addWidget(message)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)


import sys
sys.path.append('..')

try:
    from MJOLNIRGui.src.main.python.DataModels import DataSetModel,SelectionModel,DataFileModel,DataFileInfoModel,settings
    from MJOLNIRGui.src.main.python.MJOLNIR_Data import GuiDataFile,GuiDataSet
    from MJOLNIRGui.src.main.python._tools import ProgressBarDecoratorArguments,loadUI
except ImportError:
    from DataModels import DataSetModel,SelectionModel,DataFileModel,DataFileInfoModel,settings
    from MJOLNIR_Data import GuiDataFile,GuiDataSet
    from _tools import ProgressBarDecoratorArguments,loadUI

from PyQt5 import QtWidgets, QtCore, QtGui
from PyQt5.Qt import QApplication
from MJOLNIR.Data import DataFile
import numpy as np
from os import path

def setupDataSet(self): # Set up main features for Gui regarding the dataset widgets
    self.ui.DataSet_convertData_button.clicked.connect(self.DataSet_convertData_button_function)
    self.ui.DataSet_convertData_button.setToolTip('Convert selected Dataset')
    self.ui.DataSet_convertData_button.setStatusTip(self.ui.DataSet_convertData_button.toolTip())

    self.ui.DataSet_NewDataSet_button.clicked.connect(self.DataSet_NewDataSet_button_function)
    self.ui.DataSet_NewDataSet_button.setShortcut("Ctrl+N")
    self.ui.DataSet_NewDataSet_button.setToolTip('Add new Dataset')
    self.ui.DataSet_NewDataSet_button.setStatusTip(self.ui.DataSet_NewDataSet_button.toolTip())

    self.ui.DataSet_DeleteDataSet_button.clicked.connect(self.DataSet_DeleteDataSet_button_function)
    self.ui.DataSet_DeleteDataSet_button.setToolTip('Delete selected Dataset')
    self.ui.DataSet_DeleteDataSet_button.setStatusTip(self.ui.DataSet_DeleteDataSet_button.toolTip())

    self.ui.DataSet_AddFiles_button.clicked.connect(self.DataSet_AddFiles_button_function)
    self.ui.DataSet_AddFiles_button.setShortcut("Ctrl+F")
    self.ui.DataSet_AddFiles_button.setToolTip('Add new Datafiles')
    self.ui.DataSet_AddFiles_button.setStatusTip(self.ui.DataSet_AddFiles_button.toolTip())

    self.DataSetModel = DataSetModel(dataSets=self.dataSets,DataSet_DataSets_listView=self.ui.DataSet_DataSets_listView,guiWindow=self)
    self.ui.DataSet_DataSets_listView.setModel(self.DataSetModel)
    self.ui.DataSet_DataSets_listView.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)
    self.ui.DataSet_DataSets_listView.setDragDropOverwriteMode(False)

    self.DataSetSelectionModel = SelectionModel(self.DataSetModel)#self.ui.DataSet_DataSets_listView.selectionModel()
    self.DataSetModel.dragDropFinished.connect(self.DataSetSelectionModel.onModelItemsReordered)
    self.DataSetSelectionModel.selectionChanged.connect(self.selectedDataSetChanged)
    self.ui.DataSet_DataSets_listView.setSelectionModel(self.DataSetSelectionModel)
    
    def checkUpdatedName(self,index,*args):
        ds = self.model().data(index)
        suggestedName = self.model().generateValidName(ds)
        if not suggestedName == ds.name:
            ds.name = suggestedName
            self.model().layoutChanged.emit()

    self.ui.DataSet_DataSets_listView.dataChanged = checkUpdatedName

    def deleteFunction(self,gui,idx):
        gui.DataSetModel.delete(idx[0])
        self.DataFileModel.layoutChanged.emit()
        self.stateMachine.run()

    def contextMenu(view,event,gui):
        # Generate a context menu that opens on right click
        position = event.globalPos()
        idx = view.selectedIndexes()
        if len(idx)!=0:
            if event.type() == QtCore.QEvent.ContextMenu:
                menu = QtWidgets.QMenu()
                delete = QtWidgets.QAction('Delete')
                delete.setToolTip('Delete DataSet') 
                delete.setStatusTip(delete.toolTip())
                delete.triggered.connect(lambda: deleteFunction(self,gui,idx))
                delete.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/cross-button.png')))
                menu.addAction(delete)

                ds = gui.DataSetModel.data(idx[0],role = QtCore.Qt.ItemDataRole)
                if len(ds)>0:
                    currentDS = gui.DataSetModel.data(idx[0],role = QtCore.Qt.ItemDataRole)
                    if len(currentDS)!=0: # If it contains data files
                        if currentDS[0].instrument=='CAMEA': ## Only CAMEA data can be recallibrated
                            recalibrate = QtWidgets.QAction('Update Calibration')
                            recalibrate.setToolTip("Update calibration file(s)")
                            recalibrate.setStatusTip(recalibrate.toolTip())
                            recalibrate.triggered.connect(lambda: gui.DataSet_recalibrateFunction(gui,idx))
                            recalibrate.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/blue-document-resize.png')))
                            
                            menu.addAction(recalibrate)

                        sort = QtWidgets.QAction('Auto Sort')
                        sort.setToolTip("Sort data files according to ascending Ei, abs(2theta), scan dir, A3")
                        sort.setStatusTip(sort.toolTip())
                        def sorting(ds,gui):
                            ds.autoSort()
                            gui.DataSetModel.layoutChanged.emit()
                            gui.DataFileModel.layoutChanged.emit()
                        sort.triggered.connect(lambda: sorting(ds,gui))
                        sort.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/arrow-circle-double-135.png')))
                        
                        menu.addAction(sort)
                    if not ds._maskingObject is None: # A masking has been performed!
                        removeMasking = QtWidgets.QAction('Remove Masking')
                        removeMasking.setToolTip("Remove masking from dataset")
                        def removeMaskingFunction(gui,ds):
                            ds.removeMasking()
                            gui.DataSetModel.layoutChanged.emit()
                        removeMasking.setStatusTip(removeMasking.toolTip())
                        removeMasking.triggered.connect(lambda: removeMaskingFunction(gui,ds))
                        removeMasking.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/mask-cross.png')))
                        menu.addAction(removeMasking)

                return menu.exec_(position)

    self.ui.DataSet_DataSets_listView.contextMenuEvent = lambda event: contextMenu(self.ui.DataSet_DataSets_listView,event,self)

def setupDataFile(self): # Set up main features for Gui regarding the datafile widgets
    self.DataFileModel = DataFileModel(DataSet_filenames_listView=self.ui.DataSet_filenames_listView,dataSetModel=self.DataSetModel,DataSet_DataSets_listView=self.ui.DataSet_DataSets_listView,guiWindow=self)
    self.ui.DataSet_filenames_listView.setModel(self.DataFileModel)

    self.ui.DataSet_filenames_listView.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)
    self.ui.DataSet_filenames_listView.setDragDropOverwriteMode(False)

    #self.DataFileSelectionModel = self.ui.DataSet_filenames_listView.selectionModel()
    self.DataFileSelectionModel = SelectionModel(self.DataFileModel)#self.ui.DataSet_DataSets_listView.selectionModel()
    self.DataFileModel.dragDropFinished.connect(self.DataFileSelectionModel.onModelItemsReordered)
    self.DataFileSelectionModel.selectionChanged.connect(self.selectedDataFileChanged)
    self.ui.DataSet_filenames_listView.setSelectionModel(self.DataFileSelectionModel)
    
    self.ui.DataSet_DeleteFiles_button.clicked.connect(self.DataSet_DeleteFiles_button_function)
    self.ui.DataSet_DeleteFiles_button.setToolTip('Delete selected Datafile')
    self.ui.DataSet_DeleteFiles_button.setStatusTip(self.ui.DataSet_DeleteFiles_button.toolTip())

    

    self.DataFileInfoModel = DataFileInfoModel(DataSet_filenames_listView=self.ui.DataSet_filenames_listView,dataSetModel=self.DataSetModel,
    DataSet_DataSets_listView=self.ui.DataSet_DataSets_listView,dataFileModel=self.DataFileModel,guiWindow=self)
    self.setupDataFileInfoModel()
    self.ui.DataSet_fileAttributs_listView.setModel(self.DataFileInfoModel)

    def contextMenuDataFiles(view,event,gui):
        # Generate a context menu that opens on right click
        position = event.globalPos()
        idx = view.selectedIndexes()
        if len(idx)!=0:
            if event.type() == QtCore.QEvent.ContextMenu:
                menu = QtWidgets.QMenu()
                delete = QtWidgets.QAction('Delete')
                delete.setToolTip('Delete DataFile') 
                delete.setStatusTip(delete.toolTip())
                delete.triggered.connect(self.DataSet_DeleteFiles_button_function)
                delete.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/cross-button.png')))

                menu.addAction(delete)
                return menu.exec_(position)

    def dragEnterEvent(self, gui, event):
        gui.stateMachine.requireStateByName('Partial')
        if event.mimeData().hasUrls():
            # returns (path and file name, '.'+extension)
            check = [path.splitext(u.toLocalFile())[1][1:] in DataFile.supportedRawFormats for u in event.mimeData().urls()]
            if np.all(check):
                event.accept()
            else:
                event.ignore()
        else:
            event.ignore()

    def dropEvent(self, gui, event):
        files = [u.toLocalFile() for u in event.mimeData().urls()]
        gui._tempFiles = files
        gui.DataSet_AddFiles_Decorated()
        

    self.ui.DataSet_filenames_listView.dragEnterEvent = lambda event: dragEnterEvent(self.ui.DataSet_filenames_listView.dragEnterEvent,self,event)
    self.ui.DataSet_filenames_listView.dropEvent = lambda event: dropEvent(self.ui.DataSet_filenames_listView.dragEnterEvent,self,event)

    self.ui.DataSet_filenames_listView.contextMenuEvent = lambda event: contextMenuDataFiles(self.ui.DataSet_filenames_listView,event,self)
    

def setupDataSet_binning_comboBox(self):
    self.ui.DataSet_binning_comboBox.reset = lambda:DataSet_binning_comboBoxReset(self.ui.DataSet_binning_comboBox)
    self.ui.DataSet_binning_comboBox.reset()
    self.ui.DataSet_binning_comboBox.activated.connect(self.DataSet_binning_comboBox_Changed)

def selectedDataSetChanged(self,*args,**kwargs):
    self.DataFileModel.updateCurrentDataSetIndex()
    self.selectedDataFileChanged()
    self.stateMachine.run()

def selectedDataFileChanged(self,*args,**kwargs):
    self.DataFileModel.layoutChanged.emit()
    self.updateDataFileLabels()
    self.stateMachine.run()

def DataSet_NewDataSet_button_function(self):
    ds = GuiDataSet(name='Added')
    #ds.currentNormalizationSettings.update(normalizationParams)

    self.DataSetModel.append(ds)
    idx = self.DataSetModel.getCurrentDatasetIndex()
    self.ui.DataSet_DataSets_listView.edit(idx)
    self.update()
    self.stateMachine.run()

def DataSet_DeleteDataSet_button_function(self):
    idx = self.ui.DataSet_DataSets_listView.selectedIndexes()[0]
    self.DataSetModel.delete(idx)
    self.DataSetModel.layoutChanged.emit()
    self.DataFileModel.layoutChanged.emit()
    self.stateMachine.run()
    
def DataSet_DeleteFiles_button_function(self):
    self.DataFileModel.delete()
    self.stateMachine.run()

@ProgressBarDecoratorArguments(runningText='Adding Data Files',completedText='Data Files Added')
def DataSet_AddFiles_button_function(self):
    if not self.stateMachine.requireStateByName('Partial'):
        return False
    folder = self.currentFolder
    ## generate string for raw, converted, and all files
    allowedRawFilesString = 'Raw ('+' '.join(['*.'+str(x) for x in DataFile.supportedRawFormats])+')'
    allowedConvertedFilesString = 'Converted ('+' '.join(['*.'+str(x) for x in DataFile.supportedConvertedFormats])+')'
    allowedAllFilesString = 'All Files (*)'
    
    allowedString = ';;'.join([allowedRawFilesString,allowedConvertedFilesString,allowedAllFilesString])
    files, _ = QtWidgets.QFileDialog.getOpenFileNames(self,"Open data Files", folder,allowedString)
    return self.DataSet_AddFiles(files)

def DataSet_AddFiles(self,files):
    if self.DataSetModel.getCurrentDatasetIndex() is None: # no dataset is currently selected
        self.DataSet_NewDataSet_button_function()
    self.DataFileModel.add(files,guiWindow=self)

    # Find the folder of the data files, using last data file
    if len(files)==0:
        return False
    folder = path.dirname(files[-1])
    self.setCurrentDirectory(folder)

    self.update()
    self.stateMachine.run()
    return True

@ProgressBarDecoratorArguments(runningText='Converting data files',completedText='Conversion Done')
def DataSet_AddFiles_Decorated(self):
    return self.DataSet_AddFiles(self._tempFiles)

@ProgressBarDecoratorArguments(runningText='Converting data files',completedText='Conversion Done')
def DataSet_convertData_button_function(self):    
    #  Should add a check if a data set is selected
    if not self.stateMachine.requireStateByName('Raw'):
        return False
    
    val = self.convert()
    self.DataFileModel.layoutChanged.emit()
    self.DataFileInfoModel.layoutChanged.emit()
    return val
    

def recalibrateFunction(gui,idx):
    folder = gui.currentFolder
    calibrationFiles, _ = QtWidgets.QFileDialog.getOpenFileNames(gui,"Open Calibration File(s)", folder,"calbiration (*.calib);;All Files (*)")
    
    # Find the folder of the data files, using last data file
    if len(calibrationFiles)==0:
        return False
    folder = path.dirname(calibrationFiles[-1])
    gui.setCurrentDirectory(folder)

    ds = gui.DataSetModel.data(idx[0],role = QtCore.Qt.ItemDataRole)
    
    length = len(ds)+1 # Files + 1 for loading of calibration files
    gui.setProgressBarMaximum(length)
    progress = 0
    gui.setProgressBarValue(progress)
    gui.setProgressBarLabelText('Loading calibration file'+'s'*(len(calibrationFiles)>1))
    ds.updateCalibration(calibrationFiles,overwrite=True)
    progress += 1
    gui.setProgressBarValue(progress)

    files = []
    dataSetName = ds.name
    gui.setProgressBarLabelText('Updating calibration')
    for d in ds:
        files.append(d.fileLocation)
        d.saveHDF(d.fileLocation)
        progress += 1
        gui.setProgressBarValue(progress)

    gui.DataSetModel.delete(idx[0])

    ds = GuiDataSet(name=dataSetName)
    gui.setProgressBarLabelText('Reloading data file'+'s'*(len(files)>1))
    gui.DataSetModel.append(ds)
    gui.DataFileModel.add(files,guiWindow=gui)
    gui.setProgressBarLabelText('Calibration Updated')
    gui.update()
    gui.resetProgressBarTimed()
    return True
    
def convert(self):
    ds = self.DataSetModel.getCurrentDataSet()
    
    try:
        ds.convertDataFile(guiWindow=self,printFunction=self.writeToStatus)
    except AttributeError as e:
        dialog = QtWidgets.QMessageBox()
        dialog.setIcon(QtWidgets.QMessageBox.Critical)
        dialog.setText('It is not possible to convert data file:')
        
        dialog.setInformativeText(str(e))
        dialog.addButton(QtWidgets.QMessageBox.Ok)
        
        dialog.exec_()
    
    self.DataSetModel.layoutChanged.emit()
    self.DataFileModel.layoutChanged.emit()
    self.stateMachine.run()
    return True
    


def updateDataFileLabels(self):
    self.DataFileInfoModel.layoutChanged.emit()
    self.updateRaw1DCutSpinBoxes()
    self.updateBinningComboBox()

def DataSet_binning_comboBox_Changed(self):
    idx = self.ui.DataSet_binning_comboBox.currentIndex()
    newValue = int(self.ui.DataSet_binning_comboBox.itemText(idx))
    
    dfs = self.DataFileModel.getCurrentDatafiles(raw=True)
    if dfs  is None:
        dfs  = self.DataFileModel.getData() #If all or non are chosen, update all
    
    for df in dfs:
        if newValue == df.binning:
            continue

        df.binning = newValue
    self.updateDataFileLabels()

def updateBinningComboBox(self):
    dfs = self.DataFileModel.getCurrentDatafiles(raw=True)
    self.ui.DataSet_binning_comboBox.reset()
    if dfs is None:
        dfs = self.DataFileModel.getData()
    if len(dfs)==0:
        return
    possibleBinnings = set(dfs[0].possibleBinnings.copy())
    if len(dfs)>1:
        for df in dfs[1:]:
            possibleBinnings = possibleBinnings & set(df.possibleBinnings)

    possibleBinnings = list(np.sort(list(possibleBinnings)))
    self.ui.DataSet_binning_comboBox.addItems([str(b) for b in possibleBinnings])
    binning = dfs[0].binning
    idx = possibleBinnings.index(binning)
    
    self.ui.DataSet_binning_comboBox.setCurrentIndex(idx)

def DataSet_binning_comboBoxReset(self):
    values = self.count()
    for _ in range(values):
        self.removeItem(0)

############### Setup of info panel

def DataSet_DataFileInfo_copy(self,idx,clipboard=True):
    """Copy text in currently selected item to either clipboard or log"""
    
    text = self.DataFileInfoModel.data(idx,QtCore.Qt.DisplayRole)
    if clipboard:
        QApplication.clipboard().setText(text)
    else:
        self.writeToStatus(text)
    
def DataSet_DataFileInfo_remove(self,idx):
    currentInfos = self.DataFileInfoModel.infos
    del currentInfos[idx.row()]
    self.DataFileInfoModel.layoutChanged.emit()
    

def setupDataFileInfoModel(self):
    
    self.DataFileInfoModel.infos = [x for x in settings.keys()]#['sample/name','A3','A4','magneticField','temperature','scanCommand','scanParameters','comment','binning']
    def contextMenuDataFilesInfo(view,event,gui):
        # Generate a context menu that opens on right click
        
        position = event.globalPos()
        idx = view.currentIndex()
        if idx is None:
            return
        if idx.row()>=self.DataFileInfoModel.rowCount(None):
            return
    
        if event.type() == QtCore.QEvent.ContextMenu:
            menu = QtWidgets.QMenu()
            copyCB = QtWidgets.QAction('Copy to clipboard')
            copyCB.setToolTip('Copy Current information to clipboard') 
            copyCB.setStatusTip(copyCB.toolTip())
            copyCB.triggered.connect(lambda: self.DataSet_DataFileInfo_copy(idx,clipboard=True))

            copyLog = QtWidgets.QAction('Copy to log')
            copyLog.setToolTip('Copy Current information to log') 
            copyLog.setStatusTip(copyLog.toolTip())
            copyLog.triggered.connect(lambda: self.DataSet_DataFileInfo_copy(idx,clipboard=False))

            currentInfo = self.DataFileInfoModel.infos[idx.row()].baseText.split(':')[0] # base text has ': ' at the end
            removeInfo = QtWidgets.QAction("Remove '{}'".format(currentInfo))
            removeInfo.setToolTip("Remove '{}' from overview".format(currentInfo)) 
            removeInfo.setStatusTip(removeInfo.toolTip())
            removeInfo.triggered.connect(lambda: self.DataSet_DataFileInfo_remove(idx))

            

            menu.addAction(copyCB)
            menu.addAction(copyLog)
            menu.addAction(removeInfo)
            return menu.exec_(position)
    self.ui.DataSet_fileAttributs_listView.contextMenuEvent = lambda event: contextMenuDataFilesInfo(self.ui.DataSet_fileAttributs_listView,event,self)


DataSetManagerBase, DataSetManagerForm = loadUI('loadFile.ui')
       
        
class DataSetManager(DataSetManagerBase, DataSetManagerForm):
    def __init__(self, parent=None, guiWindow=None):
        super(DataSetManager, self).__init__(parent)
        self.setupUi(self)
        self.guiWindow = guiWindow
        self.initDataSetManager()

    def initDataSetManager(self):
    
        self.guiWindow.selectedDataSetChanged = lambda:selectedDataSetChanged(self.guiWindow)
        self.guiWindow.selectedDataFileChanged = lambda:selectedDataFileChanged(self.guiWindow)
        self.guiWindow.DataSet_NewDataSet_button_function = lambda:DataSet_NewDataSet_button_function(self.guiWindow)
        self.guiWindow.DataSet_DeleteDataSet_button_function = lambda:DataSet_DeleteDataSet_button_function(self.guiWindow)
        self.guiWindow.DataSet_DeleteFiles_button_function = lambda:DataSet_DeleteFiles_button_function(self.guiWindow)

        self.guiWindow.DataSet_AddFiles_button_function = lambda: DataSet_AddFiles_button_function(self.guiWindow)
        self.guiWindow.DataSet_AddFiles = lambda files: DataSet_AddFiles(self.guiWindow,files)
        self.guiWindow.DataSet_recalibrateFunction = lambda gui,idx: recalibrateFunction(gui,idx=idx)
        self.guiWindow.DataSet_AddFiles_Decorated = lambda : DataSet_AddFiles_Decorated(self.guiWindow)

        self.guiWindow.DataSet_convertData_button_function = lambda: DataSet_convertData_button_function(self.guiWindow)
        self.guiWindow.convert = lambda: convert(self.guiWindow)

        self.guiWindow.updateDataFileLabels = lambda: updateDataFileLabels(self.guiWindow)

        self.guiWindow.setupDataSet = lambda:setupDataSet(self.guiWindow)
        self.guiWindow.setupDataFile =  lambda:setupDataFile(self.guiWindow)
        self.guiWindow.setupDataFileInfoModel = lambda:setupDataFileInfoModel(self.guiWindow)
        self.guiWindow.DataSet_DataFileInfo_copy = lambda idx,clipboard:DataSet_DataFileInfo_copy(self.guiWindow,idx,clipboard)
        self.guiWindow.DataSet_DataFileInfo_remove = lambda idx: DataSet_DataFileInfo_remove(self.guiWindow, idx)
        self.guiWindow.setupDataSet_binning_comboBox = lambda:setupDataSet_binning_comboBox(self.guiWindow)
        self.guiWindow.updateBinningComboBox = lambda: updateBinningComboBox(self.guiWindow)
        self.guiWindow.DataSet_binning_comboBox_Changed = lambda:DataSet_binning_comboBox_Changed(self.guiWindow)
        
        
        
        for key,value in self.__dict__.items():
            if 'DataSet' in key:
                self.guiWindow.ui.__dict__[key] = value



    def setup(self):
        self.guiWindow.setupDataSet() # Setup datasets with buttons and call functions
        self.guiWindow.setupDataFile() # Setup datafiles      
        self.guiWindow.setupRaw1DCutSpinBoxes()
        self.guiWindow.setupDataSet_binning_comboBox()

        self.guiWindow.mask_changed.connect(self.testCall)

    @QtCore.pyqtSlot()
    def testCall(self):
        mask = self.guiWindow.maskingManager.getMasks()
        print('There was a change to the masking and I received a signal with the new mask ',mask)
        currentDS = self.guiWindow.DataSetModel.getCurrentDataSet()
        if not currentDS is None:
            currentDS.mask = mask


# State machine and corresponding classes to deal with gui movements

import numpy as np

class State(object):
    def __init__(self,name,nextState,enterStateFunction=None,transitionRequirement=None,transitionFunction=None,):
        """Create a state
        Args:
            - name (str): Name of state
            - nextState (State): Next state 
        Kwargs:
            - enterStateFunction (function): Function called by StateMachine as fun(self) (default lambda self: None)
            - transitionRequirement (function): Function returning possibility of transitioning to next state called by StateMachine as fun(self) (default lambda: False) 
            - transitionFunction (function): Function to be called to force transition to next state called by StateMachine as fun(self) (default None -> Raises error)
            
        """
        self.name = name

        if transitionRequirement is None:
            transitionRequirement = lambda self: False
        
        if enterStateFunction is None:
            enterStateFunction = lambda self: None

        self.enterStateFunction = enterStateFunction
        self.transitionRequirement = transitionRequirement
        self.nextState = nextState
        self.transitionFunction = transitionFunction




class StateMachine(object):
    def __init__(self,states,guiWindow=None):

        self.states = states
        self.stateNames = [s.name for s in self.states]
        self.guiWindow = guiWindow
        self.currentState = states[0]

    def run(self):
        """Run the machine until it cannot go any further"""
        self.currentState = self.states[0]
        while self.checkTransitionToNextState() == True: # transition requirement is made
            self.transition()
        
        self.currentState.enterStateFunction(self)
        self.guiWindow.update()
    
    def forceRun(self,checkFunction=lambda self:True):
        """Force machine to run until checkFunction returns True or next state is None"""

        while not (self.currentState.nextState is None) and (not checkFunction(self)):
            if self.checkTransitionToNextState() == False:
                if not self.currentState.transitionFunction(self):
                    return checkFunction(self)
            self.transition()
            
        self.currentState.enterStateFunction(self)
        self.guiWindow.update()
        return checkFunction(self)

    def requireStateByName(self,name):
        """Force run state machine until state with 'name' is reached"""
        #resquestIdx = self.stateNames.index(name)
        #currentIdx = self.states.index(self.currentState)
        #if currentIdx<resquestIdx:
        if not name in self.stateNames:
            raise AttributeError('Wanted name is not in list of state names! These are\n','\n'.join(self.stateNames))
        self.currentState = self.states[0]
        def checkName(self,Name):
            return self.currentState.name == Name

        check = lambda self: checkName(self,name)
        return self.forceRun(check)


    def step(self):
        if self.checkTransitionToNextState():
            self.transition()
            self.currentState.enterStateFunction(self)
            self.guiWindow.update()
            
        
    def transition(self):
        if self.checkTransitionToNextState():
            self.guiWindow.state_changed.emit(self.currentState.name,self.currentState.nextState.name)
            nextState = self.currentState.nextState
            if nextState is not None:
                self.currentState = nextState
                

    def getCurrentState(self):
        return self.currentState

    def checkTransitionToNextState(self):
        return self.currentState.transitionRequirement(self)

    @property
    def currentState(self):
        return self._currentState

    @currentState.getter
    def currentState(self):
        return self._currentState

    @currentState.setter
    def currentState(self,newState):
        if not isinstance(newState,State):
            raise AttributeError('Provided attribute is not a State')
        self._currentState = newState

        


def AllFalse(Buttons):
    return list(np.zeros_like(Buttons,dtype=bool))

def AllTrue(Buttons):
    return list(np.ones_like(Buttons,dtype=bool))

def AllFalseBut(names,buttons,buttonNames):
    buttons = AllFalse(buttons)
    for name in names:
        idx = buttonNames.index(name)
        buttons[idx] = True
    return buttons

def AllTrueBut(names,buttons,buttonNames):
    buttons = AllTrue(buttons)
    for name in names:
        idx = buttonNames.index(name)
        buttons[idx] = False
    return buttons



if __name__ == '__main__':
    ds = []
    df = []
    cf = []

    def EmptyStateInit(self):
        self.buttons = AllFalseBut(['AddDataset','AddDataFile'],self.buttons,self.buttonNames)

    def PartialStateInit(self):
        self.buttons = AllFalseBut(['AddDataset','AddDataFile'],self.buttons,self.buttonNames)

    def RawStateInit(self):
        self.buttons = AllTrueBut(['Plot'],self.buttons,self.buttonNames)

    def ConvertedStateInit(self):
        self.buttons = AllTrue(self.buttons)


    def ConvertedRequirement(ds,df,cf):
        return np.all([len(ds)!=0,len(df)!=0,len(cf)!=0])

    def RawRequirement(ds,df):
        return np.all([len(ds)!=0,len(df)!=0])

    def PartialRequirement(ds):
        return len(ds)!=0

    def addToList(ds):
        ds.append(0)
    
        

    converted = State('Converted',nextState=None,enterStateFunction=ConvertedStateInit)

    raw = State('Raw',enterStateFunction=RawStateInit,transitionRequirement=lambda:ConvertedRequirement(ds=ds,df=df,cf=cf),
                    transitionFunction=lambda:addToList(cf),nextState=converted)
    partial = State('Partial',enterStateFunction=PartialStateInit,transitionRequirement=lambda:RawRequirement(ds=ds,df=df),
                    transitionFunction=lambda:addToList(df),nextState=raw)

    empty = State('Empty',enterStateFunction=EmptyStateInit,transitionRequirement=lambda:PartialRequirement(ds=ds),
                    transitionFunction=lambda:addToList(ds),nextState=partial)



    SM = StateMachine(empty)


import sys
sys.path.append('..')
sys.path.append('.')

try:
    from Views.CalculatorGeneralManager import CalculatorGeneralManager
    from Views.CalculatorSimpleManager import CalculatorSimpleManager
    from Views.CalculatorAdvancedManager import CalculatorAdvancedManager
    from Views.collapsibleBox import CollapsibleBox
    from _tools import loadUI
except ModuleNotFoundError:
    from MJOLNIRGui.src.main.python.Views.CalculatorGeneralManager import CalculatorGeneralManager
    from MJOLNIRGui.src.main.python.Views.CalculatorSimpleManager import CalculatorSimpleManager
    from MJOLNIRGui.src.main.python.Views.CalculatorAdvancedManager import CalculatorAdvancedManager   
    from MJOLNIRGui.src.main.python.Views.collapsibleBox import CollapsibleBox
    from MJOLNIRGui.src.main.python._tools import loadUI


from os import path

from PyQt5 import QtWidgets,QtGui,QtCore



CalculatorManagerBase, CalculatorManagerForm = loadUI('Calculator.ui')

# All of this connects the buttons and their functions to the main window.
       


class CalculatorManager(CalculatorManagerBase, CalculatorManagerForm):
    def __init__(self,parent=None,guiWindow=None):

        super(CalculatorManager, self).__init__(parent=parent)
        self.setupUi(self)
        self.guiWindow = guiWindow
        self.setWindowIcon(QtGui.QIcon(self.guiWindow.AppContext.get_resource('Icons/Own/calculator.png')))

        self.nameList = ['General','Simple','Advanced'] 
        self.viewClasses = [CalculatorGeneralManager,CalculatorSimpleManager,CalculatorAdvancedManager]
        self.startState = [True,True,True] # If extended 

        self.views = []

        # Find correct layout to insert views
        vlay = self.collapsibleContainer#QtWidgets.QVBoxLayout(self.ui.collapsibleContainer)
        # Insert all views
        self.boxContainers = []
        for name,Type,state in zip(self.nameList,self.viewClasses,self.startState):
            self.update()
            box = CollapsibleBox(name,startState=state)
            self.boxContainers.append(box)
            vlay.addWidget(box)
            lay = QtWidgets.QVBoxLayout()

            widget = Type(guiWindow=self.guiWindow)
            #if Type == NormalizationManager: # Get a reference to the sample manager directly in self
            #    self.normalizationManager = widget
            self.views.append(widget)
            lay.addWidget(widget)
           
            box.setContentLayout(lay)
        vlay.addStretch()

    


from PyQt5 import QtWidgets, QtGui, QtCore
#from qtmodern.windows import ModernDialog
import os

class HelpDialog(QtWidgets.QDialog):

    def __init__(self, helpFile, guiWindow=None,*args, **kwargs):
        super(HelpDialog, self).__init__(*args, **kwargs)
        self.guiWindow = guiWindow
        if not guiWindow is None:
            self.setWindowIcon(QtGui.QIcon(self.guiWindow.AppContext.get_resource('Icons/Own/question-button.png')))
        self.setWindowTitle("Help")
        # Remove the "?" in title
        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint)
        
        

        with open(helpFile) as f:
            text = "<br>".join([line.replace('\n','') for line in f.readlines()[1:]])
            startImg = 0
            startStr = '<img src=' # look for this string

            while True:
                startImg = text.find(startStr,startImg+1)
                if startImg == -1:
                    break
                endImg = text.find('>',startImg)
                pathStart = startImg+len(startStr)
                pathNotation = text[startImg+len(startStr)] # either ' or "
                pathEnd = text.find(pathNotation,pathStart+1)
                path = text[pathStart+1:pathEnd]
                path = os.path.join(*path.split('\\'))
                text = text[:pathStart+1] + self.guiWindow.AppContext.get_resource(path) + text[pathEnd:]
            

        self.help_label = QtWidgets.QLabel(text=text)
        self.help_label.setWordWrap(True)
        self.help_label.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse|QtCore.Qt.TextSelectableByKeyboard|
                                                QtCore.Qt.LinksAccessibleByKeyboard|QtCore.Qt.LinksAccessibleByMouse)
        
        self.help_label.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.help_label.setOpenExternalLinks(True)
        
        self.layout = QtWidgets.QVBoxLayout()
        self.scroll_layout = QtWidgets.QScrollArea()
        self.scroll_layout.setLineWidth(0)
        self.scroll_layout.setFrameShape(0)
        self.layout.addWidget(self.scroll_layout)
        
        self.scroll_layout.setWidget(self.help_label)
        self.setLayout(self.layout)
        self.setMinimumSize(self.sizeHint())
        # find size of window needed
        width = max(self.help_label.size().width()+50,self.sizeHint().width())
        height = min(self.help_label.size().height(),800)
        
        self.resize(width,height)


import sys,os

from MJOLNIR import TasUBlibDEG
sys.path.append('..')

try:
    from MJOLNIRGui.src.main.python._tools import ProgressBarDecoratorArguments,loadUI
    import MJOLNIRGui.src.main.python._tools as _GUItools
except ImportError:
    from _tools import ProgressBarDecoratorArguments,loadUI
    import _tools as _GUItools
from os import path
from PyQt5 import QtWidgets,uic,QtGui,QtCore
import numpy as np
from MJOLNIR.TasUBlibDEG import calTwoTheta,calculateBMatrix,calcCell
from MJOLNIR.Geometry.Instrument import prediction,converterToA3A4,converterToQxQy,predictionInstrumentSupport
from MJOLNIR.Data import Sample
import MJOLNIR 
import pyperclip

# Handles all functionality related to the PredictionToolManager. 


PredictionToolManagerBase, PredictionToolManagerForm = loadUI('Prediction.ui')

# All of this connects the buttons and their functions to the main window.

def FocusOut(self,event,manager):
    """Custom function to overwrite default FocusOut event"""
    R = [manager.getAlignment(x) for x in [1,2]]
    if not np.all([manager.validateR(r) for r in R]): # If not valid
       manager.notValidated()
    else:
        manager.validated()
    self.focusOutEvent_old(event)


def textChangedA4Calc(alignment,manager):
    if not isinstance(alignment,list):
        alignment = [alignment]
    for a in alignment:
        A4 = manager.calculateA4(alignment=a)
        if not A4 is None:
            getattr(manager,'alignment{}_a4_spinBox'.format(a)).setValue(A4)

    manager.updateSample()
    
def onFocus(self,event,others):
    for o in others:
        try:
            o.valueChanged.disconnect()
        except TypeError: # If no connection to remove
            try:
                o.currentIndexChanged.disconnect()
            except (TypeError,AttributeError):
                pass
    if hasattr(self,'onChangeFunction'):
        try:
            self.valueChanged.connect(self.onChangeFunction)
        except AttributeError:
            self.currentIndexChanged.connect(self.onChangeFunction)
    #   self.old_focusInEvent(event)


class PredictionToolManager(PredictionToolManagerBase, PredictionToolManagerForm):
    def __init__(self, parent=None, guiWindow=None):
        super(PredictionToolManager, self).__init__(parent)
        self.setupUi(self)
        self.guiWindow = guiWindow
        self.setWindowIcon(QtGui.QIcon(self.guiWindow.AppContext.get_resource('Icons/Own/predict.png')))

        self.a4Validator = QtGui.QRegExpValidator()
        self.predictionAX = None
        if not hasattr(self.guiWindow,'braggPoints'):
            self.guiWindow.braggPoints = None
       
        regExp = QtCore.QRegExp(r'(-?[0-9]*\.[0-9]+|-?[0-9]+)(,(-?[0-9]*\.[0-9]+|-?[0-9]+))*')
        self.a4Validator.setRegExp(regExp)


        self.initPredictionToolManager()
        
    def initPredictionToolManager(self):    
        self.loadEnergies()
        self.loadSettings()
        self.setup()
        self.generateScanCommands()
        self.updateSample()


    def setup(self):

        ## Add supported instruments to combobox
        self.instrument_comboBox.addItems(predictionInstrumentSupport)

        
        # Update all boxes with check on out focus

        self.cell_a_spinBox.valueChanged.connect(lambda: textChangedA4Calc([1,2],self))
        self.cell_b_spinBox.valueChanged.connect(lambda: textChangedA4Calc([1,2],self))
        self.cell_c_spinBox.valueChanged.connect(lambda: textChangedA4Calc([1,2],self))
        self.cell_alpha_spinBox.valueChanged.connect(lambda: textChangedA4Calc([1,2],self))
        self.cell_beta_spinBox.valueChanged.connect(lambda: textChangedA4Calc([1,2],self))
        self.cell_gamma_spinBox.valueChanged.connect(lambda: textChangedA4Calc([1,2],self))

        self.alignment1_ei_spinBox.valueChanged.connect(lambda: textChangedA4Calc(1,self))
        self.alignment1_ef_comboBox.currentIndexChanged.connect(lambda Index: textChangedA4Calc(1,self))
        self.alignment1_h_spinBox.valueChanged.connect(lambda: textChangedA4Calc(1,self))
        self.alignment1_k_spinBox.valueChanged.connect(lambda: textChangedA4Calc(1,self))
        self.alignment1_l_spinBox.valueChanged.connect(lambda: textChangedA4Calc(1,self))
        self.alignment1_a3_spinBox.valueChanged.connect(lambda: textChangedA4Calc(1,self))
        
        self.alignment2_ei_spinBox.valueChanged.connect(lambda: textChangedA4Calc(2,self))
        self.alignment2_ef_comboBox.currentIndexChanged.connect(lambda Index: textChangedA4Calc(2,self))
        self.alignment2_h_spinBox.valueChanged.connect(lambda: textChangedA4Calc(2,self))
        self.alignment2_k_spinBox.valueChanged.connect(lambda: textChangedA4Calc(2,self))
        self.alignment2_l_spinBox.valueChanged.connect(lambda: textChangedA4Calc(2,self))
        self.alignment2_a3_spinBox.valueChanged.connect(lambda: textChangedA4Calc(2,self))

        
        self.scan_a4_lineEdit.setValidator(self.a4Validator)

        self.scan_a3Start_spinBox.valueChanged.connect(self.generateScanCommands)
        self.scan_a3Stop_spinBox.valueChanged.connect(self.generateScanCommands)
        self.scan_a3Steps_spinBox.valueChanged.connect(self.generateScanCommands)
        self.scan_ei_spinBox.valueChanged.connect(self.generateScanCommands)
        self.scan_a4_lineEdit.textChanged.connect(self.generateScanCommands)
        self.scan_monitor_spinBox.valueChanged.connect(self.generateScanCommands)

        self.scan_curratAxe_pushButton.clicked.connect(self.curratAxeList)

        self.tool_generate_button.clicked.connect(self.generatePrediction)

        self.HKL_ei_spinBox.onChangeFunction = self.calcualteHKLtoA3A4
        self.HKL_ef_comboBox.onChangeFunction = lambda I: self.calcualteHKLtoA3A4()

        self.HKL_H_doubleSpinBox.onChangeFunction = self.calcualteHKLtoA3A4
        self.HKL_K_doubleSpinBox.onChangeFunction = self.calcualteHKLtoA3A4
        self.HKL_L_doubleSpinBox.onChangeFunction = self.calcualteHKLtoA3A4

        self.HKL_A3_doubleSpinBox.onChangeFunction = self.calcualteA3A4toHKL
        self.HKL_A4_doubleSpinBox.onChangeFunction = self.calcualteA3A4toHKL


        self.HKL_A3_doubleSpinBox.old_focusInEvent = self.HKL_A3_doubleSpinBox.focusInEvent
        self.HKL_A4_doubleSpinBox.old_focusInEvent = self.HKL_A4_doubleSpinBox.focusInEvent
        self.HKL_H_doubleSpinBox.old_focusInEvent = self.HKL_H_doubleSpinBox.focusInEvent
        self.HKL_K_doubleSpinBox.old_focusInEvent = self.HKL_K_doubleSpinBox.focusInEvent
        self.HKL_L_doubleSpinBox.old_focusInEvent = self.HKL_L_doubleSpinBox.focusInEvent
        self.HKL_ei_spinBox.old_focusInEvent = self.HKL_ei_spinBox.focusInEvent
        self.HKL_ef_comboBox.old_focusInEvent = self.HKL_ef_comboBox.focusInEvent
        
        
        # Update focusing for A3 and A4
        others = [self.HKL_H_doubleSpinBox,self.HKL_K_doubleSpinBox,self.HKL_L_doubleSpinBox]
        self.HKL_A3_doubleSpinBox.focusInEvent= lambda event: onFocus(self.HKL_A3_doubleSpinBox,event,others)
        self.HKL_A4_doubleSpinBox.focusInEvent= lambda event: onFocus(self.HKL_A4_doubleSpinBox,event,others)

        # Update focusing for H,K,L,Ei,Ef
        others = [self.HKL_A3_doubleSpinBox,self.HKL_A4_doubleSpinBox]
        self.HKL_H_doubleSpinBox.focusInEvent= lambda event: onFocus(self.HKL_H_doubleSpinBox,event,others)
        self.HKL_K_doubleSpinBox.focusInEvent= lambda event: onFocus(self.HKL_K_doubleSpinBox,event,others)
        self.HKL_L_doubleSpinBox.focusInEvent= lambda event: onFocus(self.HKL_L_doubleSpinBox,event,others)

        self.HKL_ei_spinBox.focusInEvent= lambda event: onFocus(self.HKL_ei_spinBox,event,others)
        self.HKL_ef_comboBox.focusInEvent= lambda event: onFocus(self.HKL_ef_comboBox,event,others)    

        
        for key,value in self.__dict__.items():
            if hasattr(value,'valueCanged'):
                value.valueChanged.connect(self.updateSettings)
            elif hasattr(value,'textChanged'):
                value.textChanged.connect(self.updateSettings)
        
    def updateSample(self):
        cell = np.array(self.getCell())
        r1 = np.array(self.getAlignment(1))
        r2 = np.array(self.getAlignment(2))
        self.sample = Sample.Sample(*cell,projectionVector1=r1,projectionVector2=r2)
        self.calcualteHKLtoA3A4()

    def getAlignment(self,alignment=1):
        """Get values for alignment vector 1"""
        Ei = getattr(self,'alignment{}_ei_spinBox'.format(alignment)).value()
        EfIndex = getattr(self,'alignment{}_ef_comboBox'.format(alignment)).currentIndex()
        if EfIndex == len(self.Efs): # Equal to last entry => Ei=Ef
            Ef = Ei
        else:
            Ef = self.Efs[EfIndex]

        H = getattr(self,'alignment{}_h_spinBox'.format(alignment)).value()
        K = getattr(self,'alignment{}_k_spinBox'.format(alignment)).value()
        L = getattr(self,'alignment{}_l_spinBox'.format(alignment)).value()
        A3 = getattr(self,'alignment{}_a3_spinBox'.format(alignment)).value()
        A4 = getattr(self,'alignment{}_a4_spinBox'.format(alignment)).value()

        return [H,K,L,A3,A4,0.0,0.0,Ei,Ef] # H,K,L,A3,A4,phi,chi,Ei,Ef

    def setAlignment(self,R,alignment=1):
        [H,K,L,A3,A4,_,_,Ei,Ef] = R
        getattr(self,'alignment{}_ei_spinBox'.format(alignment)).setValue(Ei)
        EfIndex = np.argmin(np.abs(Ef-self.Efs))
        getattr(self,'alignment{}_ef_comboBox'.format(alignment)).setCurrentIndex(EfIndex)

        getattr(self,'alignment{}_h_spinBox'.format(alignment)).setValue(H)
        getattr(self,'alignment{}_k_spinBox'.format(alignment)).setValue(K)
        getattr(self,'alignment{}_l_spinBox'.format(alignment)).setValue(L)
        getattr(self,'alignment{}_a3_spinBox'.format(alignment)).setValue(A3)
        getattr(self,'alignment{}_a4_spinBox'.format(alignment)).setValue(A4)

    def calculateA4(self,alignment=1):
        cell = self.getCell()
        r = self.getAlignment(alignment=alignment)

        Cell = calcCell(cell)
        B = calculateBMatrix(Cell)

        # H,K,L,Ei,Ef
        qe = np.concatenate([r[:3],r[-2:]])

        A4 = calTwoTheta(B=B,qe=qe,ss=-1)
        return A4

    def loadEnergies(self,instrument='CAMEA'):
        if instrument == 'CAMEA':
            calibPath = MJOLNIR.__CAMEANormalizationBinning1__

            calib = np.loadtxt(calibPath,delimiter=',',skiprows=3)

            self.A4Instrument = calib[:,-1].reshape(104,8)
            self.EfInstrument = calib[:,4].reshape(104,8)
            self.EfInstrument[np.isclose(self.EfInstrument,0.0)]=np.nan
        else:
            raise NotImplementedError('Currently only CAMEA is created... sorry')

        self.Efs = np.nanmean(self.EfInstrument,axis=0)[::-1]
        for key,value in self.__dict__.items():
            if 'ef_comboBox' in key:
                for i,ef in enumerate(self.Efs):
                    value.addItem('{:.2f} ({:d})'.format(ef,len(self.Efs)-i-1))

                value.addItem(' = Ei')
        


    def validated(self):
        self.tool_generate_button.setDisabled(False)
        self.tool_generate_button.setStyleSheet("color: black;")

    def notValidated(self):
        self.tool_generate_button.setDisabled(True)
        self.tool_generate_button.setStyleSheet("color: red;")

    def validateR(self,r):
        """Validate if r vector is valid. If HKL is nonzero, A4 is nonzero and Energies are nonzero."""

        H,K,L,_,A4,_,_,Ei,Ef = r
        if np.isclose(np.linalg.norm([H,K,L]),0.0):
            return False
        if np.isclose(A4,0.0):
            return False
        if np.any([np.isclose(E,0.0) for E in [Ei,Ef]]):
            return False
        
        return True
    
        
    def generatePrediction(self):
        """Generate prediction window from MJOLNIR"""
        A3Start,A3Stop,A3Steps,Ei,A4,points,Monitor = self.getScan()
        r1 = np.array(self.getAlignment(alignment=1))
        r2 = np.array(self.getAlignment(alignment=2))

        cell = np.array(self.getCell())

        instrument = str(self.instrument_comboBox.currentText())

        ax = prediction(A3Start=A3Start,A3Stop=A3Stop,A3Steps=A3Steps,A4Positions=A4,Ei=Ei,Cell=cell,r1=r1,r2=r2,
        points=points,outputFunction=self.guiWindow.writeToStatus, instrument=instrument)

        self.predictionAx = ax
        self.guiWindow.windows.append(ax[0].get_figure())

        # make all figures have the same xlim and ylim by using the last axis

        for a in ax[:-1]:
            a.set_ylim(*ax[-1].get_ylim())
            a.set_xlim(*ax[-1].get_xlim())

        curratAxe = self.scan_curratAxe_checkBox.isChecked()

        if curratAxe:
            self.plotCurratAxe()
        
    def plotCurratAxe(self):
        if hasattr(self.guiWindow,'BraggListWindow'):
            BraggPoint = np.array(self.guiWindow.BraggListWindow.BraggListModel.getAllData())
        elif hasattr(self.guiWindow,'braggPoints'):
            BraggPoint = self.guiWindow.braggPoints
        else:
            return

        if not hasattr(self,'predictionAx'):
            return
        
        Ef = self.Efs[::-1]
        Ei = self.getScan()[3]
        SpurionPositionsMono = self.predictionAx[0].sample.CurratAxe(Ei=Ei,Ef=Ef,Bragg=BraggPoint,HKL=False,spurionType = 'Monochromator').reshape(len(BraggPoint),-1,3).transpose(1,0,2)
        SpurionPositionsAna  = self.predictionAx[0].sample.CurratAxe(Ei=Ei,Ef=Ef,Bragg=BraggPoint,HKL=False,spurionType = 'Analyser').reshape(len(BraggPoint),-1,3).transpose(1,0,2)
        
        
        self.monoPoints = []
        self.anaPoints = []
        for ax,posMono,posAna in zip(self.predictionAx,SpurionPositionsMono,SpurionPositionsAna):
            for bragg in posMono:
                self.monoPoints.append( ax.scatter(*bragg[:2],marker='o',color='r'))
            for bragg in posAna:
                self.anaPoints.append(ax.scatter(*bragg[:2],marker='o',color='k'))

    def removeCurratAxe(self):
        if hasattr(self,'predictionAx'):
            if hasattr(self,'monoPoints'):
                for p in np.concatenate(self.monoPoints,self.anaPoints):
                    p.setVisible(False)



    def getScan(self):
        A3Start = self.scan_a3Start_spinBox.value()
        A3Stop = self.scan_a3Stop_spinBox.value()
        A3Steps = self.scan_a3Steps_spinBox.value()
        Ei = self.scan_ei_spinBox.value()

        A4 = self.formatA4String(self.scan_a4_lineEdit.text())

        points = self.scan_plot_checkBox.isChecked()

        Monitor = self.scan_monitor_spinBox.value()

        return A3Start,A3Stop,A3Steps,Ei,A4,points,Monitor

    def setScan(self,scan):
        A3Start,A3Stop,A3Steps,Ei,A4,points,Monitor = scan
        self.scan_a3Start_spinBox.setValue(A3Start)
        self.scan_a3Stop_spinBox.setValue(A3Stop)
        self.scan_a3Steps_spinBox.setValue(A3Steps)
        self.scan_ei_spinBox.setValue(Ei)
        self.scan_monitor_spinBox.setValue(Monitor)

        strA4 = ','.join([str(a4) for a4 in A4])
        self.scan_a4_lineEdit.setText(strA4)
        
        points = self.scan_plot_checkBox.setChecked(points)

    
    def formatA4String(self,A4String):
        if ',' in A4String:
            A4 = []
            for x in A4String.split(','):
                try:
                    A4.append(float(x))
                except ValueError:
                    pass
        else:
            A4 = [float(A4String)]

        return A4

    def getCell(self):
        a = self.cell_a_spinBox.value()
        b = self.cell_b_spinBox.value()
        c = self.cell_c_spinBox.value()
        alpha = self.cell_alpha_spinBox.value()
        beta = self.cell_beta_spinBox.value()
        gamma = self.cell_gamma_spinBox.value()

        return a,b,c,alpha,beta,gamma


    def setCell(self,cell):
        a,b,c,alpha,beta,gamma = cell

        self.cell_a_spinBox.setValue(a)
        self.cell_b_spinBox.setValue(b)
        self.cell_c_spinBox.setValue(c)
        self.cell_alpha_spinBox.setValue(alpha)
        self.cell_beta_spinBox.setValue(beta)
        self.cell_gamma_spinBox.setValue(gamma)

    def updateSettings(self):
        """Update self.guiWindow.predictionSettings with current settings"""
        R1 = self.getAlignment(alignment = 1)
        R2 = self.getAlignment(alignment = 2)
        cell = self.getCell()
        scan = self.getScan()
        calc = self.getCalculation()

        names = ['R1','R2','cell','scan','calc']
        values = [R1,R2,cell,scan,calc]
        for name,value in zip(names,values):
            self.guiWindow.predictionSettings[name] = value

    def loadSettings(self):
        # Set up saving of settings in guiWindow

        if hasattr(self.guiWindow,'predictionSettings'): 
            if self.guiWindow.predictionSettings: # If not empty
                R1 = self.guiWindow.predictionSettings['R1']
                R2 = self.guiWindow.predictionSettings['R2']
                cell = self.guiWindow.predictionSettings['cell']
                scan = self.guiWindow.predictionSettings['scan']
                calc = self.guiWindow.predictionSettings['calc']
                self.setAlignment(R1,alignment = 1)
                self.setAlignment(R2,alignment = 2)
                self.setCell(cell)
                self.setScan(scan)
                self.setCalculation(calc)
            else:# Create an empty dict
                self.guiWindow.predictionSettings = {}

        else:# Create an empty dict
            self.guiWindow.predictionSettings = {}

    def generateScanCommands(self):
        
        A3Start,A3Stop,A3Steps,Ei,A4,points,Monitor = self.getScan()
        if A3Steps == 1:
            A3StepSize = 0
        else:
            A3StepSize = (A3Stop-A3Start)/(A3Steps-1.0)
        A3middle = 0.5*(A3Start+A3Stop)

        scanCommand = 'sc a3 {:.2f} da3 {:.2f} np {:d} mn {:}'.format(A3middle,A3StepSize,A3Steps,Monitor)

        commandString = []
        commandString.append('dr 2t {:.2f}'.format(A4[0]))
        commandString.append('dr ei {:.2f}'.format(Ei))
        commandString.append('')
        for a4 in A4:
            commandString.append('dr 2t {:.2f}'.format(a4))
            commandString.append(scanCommand)
            commandString.append('')

        commandString.append('')
        cmdStr = '\n'.join(commandString)
        self.scanCommand_textEdit.setText(cmdStr)

    def getCalculation(self):
        Ei = self.HKL_ei_spinBox.value()
        EfIndex = self.HKL_ef_comboBox.currentIndex()
        if EfIndex == len(self.Efs): # Equal to last entry => Ei=Ef
            Ef = Ei
        else:
            Ef = self.Efs[EfIndex]
        H = self.HKL_H_doubleSpinBox.value()
        K = self.HKL_K_doubleSpinBox.value()
        L = self.HKL_L_doubleSpinBox.value()
        A3 = self.HKL_A3_doubleSpinBox.value()
        A4 = self.HKL_A4_doubleSpinBox.value()

        return Ei,Ef,H,K,L,A3,A4,EfIndex

    def setCalculation(self,calc):
        Ei,Ef,H,K,L,A3,A4,EfIndex = calc

        self.HKL_ef_comboBox.setCurrentIndex(EfIndex)
        self.HKL_ei_spinBox.setValue(Ei)
        self.HKL_H_doubleSpinBox.setValue(H)
        self.HKL_K_doubleSpinBox.setValue(K)
        self.HKL_L_doubleSpinBox.setValue(L)
        self.HKL_A3_doubleSpinBox.setValue(A3)
        self.HKL_A4_doubleSpinBox.setValue(A4)

    def calcualteHKLtoA3A4(self):
        Ei,Ef,H,K,L,*_ = self.getCalculation()
        
        UB = self.sample.UB

        Qx,Qy,_ = np.dot(UB,[H,K,L])
        A3,A4 = converterToA3A4(Qx,Qy,Ei,Ef,A3Off=0.0,A4Sign=np.sign(self.sample.plane_vector1[4]))
        
        self.HKL_A3_doubleSpinBox.setValue(A3)
        self.HKL_A4_doubleSpinBox.setValue(A4)
        
    def calcualteA3A4toHKL(self):
        Ei,Ef,_,_,_,A3,A4,_ = self.getCalculation()
        Qx,Qy = converterToQxQy(A3,A4,Ei,Ef)
        H,K,L = self.sample.calculateQxQyToHKL(Qx,Qy)
        self.HKL_H_doubleSpinBox.setValue(H)
        self.HKL_K_doubleSpinBox.setValue(K)
        self.HKL_L_doubleSpinBox.setValue(L)

    def curratAxeList(self):
        self.guiWindow.openBraggListWindow()


    def closeEvent(self, event):
        self.updateSettings()


import sys
sys.path.append('..')

try:
#    import MJOLNIRGui.src.main.python._tools as _GUItools
    from MJOLNIRGui.src.main.python.DataModels import BraggListModel
    from MJOLNIRGui.src.main.python._tools import loadUI
except ImportError:
    from DataModels import BraggListModel
    from _tools import loadUI
#    import _tools as _GUItools
from os import path
from PyQt5 import QtWidgets, uic, QtCore, QtGui
import numpy as np



BraggListManagerBase, BraggListManagerForm = loadUI('braggListWidget.ui')

class BraggListManager(BraggListManagerBase, BraggListManagerForm):
    def __init__(self, parent=None, guiWindow=None,BraggList=None):
        super(BraggListManager, self).__init__(parent)
        self.setupUi(self)
        self.guiWindow = guiWindow
        self.initBraggList = BraggList
        self.initBraggListManager()
        self.selectedChanged()

    def initBraggListManager(self):
        # set multiple shortcuts for add_button
        for sequence in ("Ctrl+Enter", "Ctrl+Return",):
            shorcut = QtWidgets.QShortcut(sequence, self.BraggList_add_button)
            shorcut.activated.connect(self.BraggList_add_button_function)

        self.BraggList_delete_button.clicked.connect(self.BraggList_delete_button_function)
        self.BraggList_delete_button.setShortcut("Delete")
        self.BraggList_deleteAll_button.clicked.connect(self.BraggList_deleteAll_button_function)
        self.BraggList_deleteAll_button.setShortcut("Ctrl+Delete")

        self.BraggListModel = BraggListModel(braggList_listView=self.BraggList_listView,BraggList=self.initBraggList)
        self.BraggList_listView.setModel(self.BraggListModel)

        self.BraggListSelectionModel = self.BraggList_listView.selectionModel()
        self.BraggListModel.layoutChanged.connect(self.selectedChanged)
        self.BraggListSelectionModel.selectionChanged.connect(self.selectedChanged)
        
    def closeEvent(self, event): # Function called on close event for the window
        self.guiWindow.braggPoints = self.getData()

    def selectedChanged(self,*args,**kwargs):
        if self.BraggListModel.rowCount() == 0:
            self.BraggList_deleteAll_button.setDisabled(True)
            self.BraggList_listView.clearSelection()
        else:
            self.BraggList_deleteAll_button.setDisabled(False)

        if len(self.BraggList_listView.selectedIndexes()) == 0:
            self.BraggList_delete_button.setDisabled(True)
        else:
            self.BraggList_delete_button.setDisabled(False)
    
    
    def BraggList_delete_button_function(self):
        self.BraggListModel.delete(self.BraggList_listView.selectedIndexes())
        self.BraggListModel.layoutChanged.emit()

    def BraggList_add_button_function(self):
        HKL = self.extractBraggParamters()
        self.BraggListModel.append(HKL)

    def BraggList_deleteAll_button_function(self):
        self.BraggListModel.deleteAll()
        self.BraggListModel.layoutChanged.emit()

    def getData(self):
        data = self.BraggListModel.getAllData()
        if len(data) == 0:
            return None
        return data


    def extractBraggParamters(self):
        H = self.BraggList_H_spinBox.value()
        K = self.BraggList_K_spinBox.value()
        L = self.BraggList_L_spinBox.value()

        return H,K,L


from setuptools import setup
import os
import sys
import json


_here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(_here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


settings = {'app_name': 'MJOLNIRGui', 
'author': 'MJOLNIRPackage', 
'main_module': 'src/main/python/main.py', 
'version': '0.9.10',
'gpg_name': 'MJOLNIRPackage', 
'hidden_imports': ['cmath'], 
'categories': 'Education;', 
'description': 'Graphical user interface for the MJOLNIR software. Documentation and further information can be found at https://www.psi.ch/en/sinq/camea/data-treatment. Please do cite this package with the DOI and article found at the website.', 
'author_email': 'MJOLNIRPackage@gmail.dk', 
'url': 'https://github.com/MJOLNIRPackage/MJOLNIR-GUI', 
'mac_bundle_identifier': 'ch.psi.MJOLNIRGui'}


home = os.path.join(_here,'MJOLNIRGui')
packages = [x[0] for x in os.walk(home) if x[0][-1]!='_']

packages = [os.path.relpath(p,_here) for p in packages]


setup(
    name=settings['app_name'],
    version=settings['version'],
    description=('Neutron Scattering software suite.'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=settings['author'],
    author_email=settings['author_email'],
    url=settings['url'],
    license='MPL-2.0',
    packages=packages,
    package_data={'': ['*']},
    include_package_data=True,
    entry_points = {
        "console_scripts": ['MJOLNIRGui = MJOLNIRGui.src.main.python.MJOLNIR_GUI:main']
        },
    python_requires='>=3.5',
    install_requires=['pip>=20','PyQt5','MJOLNIR>=1.1.24','ufit>=1.4.0'], 
    
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        ],
    )



# -*- coding: utf-8 -*-
"""
Spyder Editor

Script to copy all dependencies into the MJOLNIRGui folder to prepare for pip
installation
"""

import os, shutil
def copytree(src, dst, symlinks=False, ignore=None, remove=None):
    if remove is None:
        remove = []
    for item in os.listdir(src):
        if os.path.split(item)[-1] in remove:
            continue
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
            print('Copying folder {} to {}'.format(s,d))
        else:
            shutil.copy2(s, d)
            print('Copying file {} to {}'.format(s,d))


delCommand = 'del'

local = os.path.dirname(__file__)

targetDir = os.path.join(local,'MJOLNIRGui','src')


if not os.path.exists(targetDir):
    os.system("mkdir {}".format(targetDir))

# If the build folder exists also delete the main folder
if os.path.exists(os.path.join(targetDir,'build')):
    shutil.rmtree(os.path.join(targetDir,'build'))
    shutil.rmtree(os.path.join(targetDir,'main'))

# Do not include the installer or sign folders
removeList = ['installer','sign']

copytree(os.path.join(local,'src'),targetDir,remove=removeList)

# Remove secret json file (Currently not present)
secretFile = os.path.join(targetDir,'build','settings','secret.json')
if os.path.exists(secretFile):
    os.remove(os.path.join(targetDir,'build','settings','secret.json'))




import sys
sys.path.append('..')

try:
    from MJOLNIRGui.src.main.python.MJOLNIR_Data import GuiDataFile,GuiDataSet
    from MJOLNIRGui.src.main.python._tools import ProgressBarDecoratorArguments,loadUI
except ImportError:
    from MJOLNIR_Data import GuiDataFile,GuiDataSet
    from _tools import ProgressBarDecoratorArguments,loadUI

from os import path
from PyQt5 import QtWidgets,uic
import numpy as np
from MJOLNIR.Data import DataFile


def setupRaw1DCutSpinBoxes(self):
    self.ui.Raw1D_Analyzer_spinBox.valueChanged.connect(self.raw1DCutAnalyzerSpinBoxChanged)
    self.ui.Raw1D_Detector_spinBox.valueChanged.connect(self.raw1DCutDetectorSpinBoxChanged)
    self.resetRaw1DCutSpinBoxes()

        
def resetRaw1DCutSpinBoxes(self):
    self.ui.Raw1D_Analyzer_spinBox.setEnabled(False)
    self.ui.Raw1D_Detector_spinBox.setEnabled(False)
    self.ui.Raw1D_Analyzer_spinBox.setValue(0)
    self.ui.Raw1D_Detector_spinBox.setValue(0)

    self.ui.Raw1D_Analyzer_Original_label.setText('Original N/A')
    self.ui.Raw1D_Detector_Original_label.setText('Original N/A')

    EfEntry = '{:.2f}'.format(0).rjust(9,' ')
    A4Entry = '{:+.2f}'.format(0).rjust(9,' ')

    self.ui.Raw1D_Analyzer_label.setText('Analyzer number (Ef = {} meV)'.format(EfEntry))
    self.ui.Raw1D_Detector_label.setText('Detector number (A4 = {} deg)'.format(A4Entry))


def updateRaw1DCutSpinBoxes(self,dfs=None):
    if dfs is None:
        dfs = self.DataFileModel.getCurrentDatafiles()
        if dfs is None:
            return self.resetRaw1DCutSpinBoxes()
    if len(dfs)>1:
        return self.resetRaw1DCutSpinBoxes()
    
    df = dfs[0]
    

    maxAnalyzer = DataFile.analyzerLimits[df.instrument]
    maxDetector = DataFile.detectorLimits[df.instrument]
    self.ui.Raw1D_Analyzer_spinBox.setEnabled(True)
    self.ui.Raw1D_Detector_spinBox.setEnabled(True)
    self.ui.Raw1D_Analyzer_spinBox.setMaximum(maxAnalyzer)
    self.ui.Raw1D_Detector_spinBox.setMaximum(maxDetector)
    self.ui.Raw1D_Analyzer_spinBox.setValue(df.analyzerSelection)
    self.ui.Raw1D_Detector_spinBox.setValue(df.detectorSelection)
    self.updateRaw1DCutLabels(dfs)
    

def updateRaw1DCutLabels(self,dfs=None):
    if dfs is None:
        dfs = self.DataFileModel.getCurrentDatafiles()
        if dfs is None:
            return self.resetRaw1DCutSpinBoxes()
    if len(dfs)>1:
        return self.resetRaw1DCutSpinBoxes()
    
    df = dfs[0]

    self.ui.Raw1D_Analyzer_Original_label.setText('Original {}'.format(df.analyzerSelectionOriginal))
    self.ui.Raw1D_Detector_Original_label.setText('Original {}'.format(df.detectorSelectionOriginal))

    binning = 1
    calibrationIndex = list(df.possibleBinnings).index(binning) # Only binning 1 is used for raw plotting
    
    if df.instrument == 'CAMEA':
        EPrDetector = 8 
        detectors = 104
    elif df.type == 'MultiFLEXX':
        EPrDetector = 5
        detectors = 31
    elif df.type == 'FlatCone':
        EPrDetector = 1
        detectors = 31
    elif df.type == 'Bambus' or df.instrument == 'Bambus':
        EPrDetector = 1
        detectors = 100
    else:
        totalDetectors = np.array(df.instrumentCalibrations[calibrationIndex][0].shape[:-1])
        if len(totalDetectors) == 2:
            detectors,EPrDetector = totalDetectors
        else:
            if np.mod(totalDetectors,31)==0: # either MultiFLEXX or FlatCone
                EPrDetector = int(totalDetectors/31)
                detectors = 31
            else: # CAMEA
                EPrDetector = 8 
                detectors = 104
    
    instrumentCalibrationEf,instrumentCalibrationA4,_ = df.instrumentCalibrations[calibrationIndex]
    
    
    instrumentCalibrationEf.shape = (-1,4)
    instrumentCalibrationA4.shape = (-1)

    analyzerValue = self.ui.Raw1D_Analyzer_spinBox.value()
    detectorValue = self.ui.Raw1D_Detector_spinBox.value() #
    idxDetector,_ = df.calcualteDataIndexFromDasel(detectorSelection=detectorValue,analyzerSelection=analyzerValue)
    Ef = instrumentCalibrationEf[idxDetector,1]
    A4 = instrumentCalibrationA4[idxDetector]

    EfEntry = '{:.2f}'.format(Ef).rjust(9,' ')
    A4Entry = '{:+.2f}'.format(A4).rjust(9,' ')

    self.ui.Raw1D_Analyzer_label.setText('Analyzer number (Ef = {} meV)'.format(EfEntry))
    self.ui.Raw1D_Detector_label.setText('Detector number (A4 = {} deg)'.format(A4Entry))

    
def raw1DCutAnalyzerSpinBoxChanged(self):
    value = self.ui.Raw1D_Analyzer_spinBox.value()
    dfs = self.DataFileModel.getCurrentDatafiles()
    if dfs is None:
        return None
    if len(dfs)>1:
        return None
    
    df = dfs[0]
    df.analyzerSelection = np.array(value)
    self.updateRaw1DCutLabels(dfs=dfs)

def raw1DCutDetectorSpinBoxChanged(self):
    value = self.ui.Raw1D_Detector_spinBox.value()
    dfs = self.DataFileModel.getCurrentDatafiles()
    if dfs is None:
        return None
    if len(dfs)>1:
        return None
    
    df = dfs[0]
    df.detectorSelection = np.array(value)
    self.updateRaw1DCutLabels(dfs=dfs)
        
@ProgressBarDecoratorArguments(runningText='Plotting raw 1D Data',completedText='Plotting Done')
def Raw1D_plot_button_function(self):
    if not self.stateMachine.requireStateByName('Raw'):
        return False
    dataFiles = self.DataFileModel.getCurrentDatafiles()
    if dataFiles is None:
        ds = self.DataSetModel.getCurrentDataSet()
    elif len(dataFiles) == 0: # No data files selected, use them all
        ds = self.DataSetModel.getCurrentDataSet()
    else:
        ds = GuiDataSet(dataFiles)
    
    if ds is None:
        return False

    
    ax = ds.plotRaw1D()
    self.windows.append(ax.get_figure())
    return True


Raw1DManagerBase, Raw1DManagerForm = loadUI('Raw1D.ui')

class Raw1DManager(Raw1DManagerBase, Raw1DManagerForm):
    def __init__(self, parent=None, guiWindow=None):
        super(Raw1DManager, self).__init__(parent)
        self.setupUi(self)
        self.guiWindow = guiWindow
        self.initRaw1DManager()


    def initRaw1DManager(self):

        self.guiWindow.setupRaw1DCutSpinBoxes = lambda: setupRaw1DCutSpinBoxes(self.guiWindow)
        self.guiWindow.resetRaw1DCutSpinBoxes = lambda: resetRaw1DCutSpinBoxes(self.guiWindow)
        self.guiWindow.updateRaw1DCutSpinBoxes = lambda dfs=None: updateRaw1DCutSpinBoxes(self.guiWindow, dfs)
        self.guiWindow.updateRaw1DCutLabels = lambda dfs=None: updateRaw1DCutLabels(self.guiWindow,dfs)
        self.guiWindow.raw1DCutAnalyzerSpinBoxChanged = lambda: raw1DCutAnalyzerSpinBoxChanged(self.guiWindow)
        
        self.guiWindow.raw1DCutDetectorSpinBoxChanged = lambda: raw1DCutDetectorSpinBoxChanged(self.guiWindow)
        self.guiWindow.Raw1D_plot_button_function = lambda: Raw1D_plot_button_function(self.guiWindow)

        for key,value in self.__dict__.items():
            if 'Raw1D' in key:
                self.guiWindow.ui.__dict__[key] = value

    def setup(self):
        
        self.guiWindow.ui.Raw1D_plot_button.clicked.connect(self.guiWindow.Raw1D_plot_button_function)
    


from re import sub
import sys
sys.path.append('..')

try:
    from MJOLNIRGui.src.main.python.DataModels import DataSetModel,SelectionModel,DataFileModel,DataFileInfoModel,subtractionSettings
    from MJOLNIRGui.src.main.python.MJOLNIR_Data import GuiDataFile,GuiDataSet
    from MJOLNIRGui.src.main.python._tools import ProgressBarDecoratorArguments,loadUI
    from MJOLNIRGui.src.main.python.HelpDialog import HelpDialog
except ImportError:
    from DataModels import DataSetModel,SelectionModel,DataFileModel,DataFileInfoModel,subtractionSettings
    from MJOLNIR_Data import GuiDataFile,GuiDataSet
    from _tools import ProgressBarDecoratorArguments,loadUI
    from HelpDialog import HelpDialog


from os import path
from PyQt5 import QtWidgets,uic, QtCore, QtGui
from PyQt5.QtCore import Qt
import numpy as np
def subtractable(file1,file2,A3Precision=0.1,twothetaPrecision=0.1,EiPrecision=0.1):
    checked = {}
    checked['type'] = file1.type == file2.type
    checked['scanSteps'] = file1.I.shape == file2.I.shape
    if not file1.A3.shape ==  file2.A3.shape:
        checked['A3'] = False
    else:
        checked['A3'] = np.all(np.isclose(file1.A3,file2.A3,atol=A3Precision))
    checked['twotheta'] = np.all(np.isclose(file1.twotheta,file2.twotheta,atol=twothetaPrecision))
    checked['Ei'] = np.all(np.isclose(file1.Ei,file2.Ei,atol=EiPrecision))
    checked['scanParameters'] = file1.scanParameters == file2.scanParameters
    return checked

ValidColour =  QtGui.QColor(255, 255, 255, 0)
InvalidColour = QtGui.QColor(255, 0, 0, 120)

class DataSetDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, parent=None, *args):
        """Delegate colouring the background of data files with the BG colour given by df.BGColor, default is ValidColor"""
        QtWidgets.QStyledItemDelegate.__init__(self, parent, *args)


    def paint(self, painter, option, index):
        super(DataSetDelegate,self).paint(painter, option, index)
        if True:
            df = index.model().data(index,Qt.ItemDataRole)
            painter.save()
            painter.setPen(QtGui.QPen(Qt.NoPen))
            
            if not hasattr(df,'BGColor'): # No BG color given, assume Valid
                c = ValidColour
            else:
                c = df.BGColor
            painter.setBrush(QtGui.QBrush(c))

            # Draw the background rectangle            
            painter.drawRect(option.rect)
            painter.restore()

class DataFileDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, parent=None, *args):
        """Delegate colouring the background of data file properties depending on whether they match or not"""
        QtWidgets.QStyledItemDelegate.__init__(self, parent, *args)


    def paint(self, painter, option, index):
        super(DataFileDelegate,self).paint(painter, option, index)
        
        dataPropertyCheck = index.model().data(index,role=Qt.ItemDataRole)
    
        painter.save()
        painter.setPen(QtGui.QPen(Qt.NoPen))
        
        if dataPropertyCheck: # Check is successful
            c = ValidColour
        else:
            c = InvalidColour
        painter.setBrush(QtGui.QBrush(c))

        # Draw the background rectangle            
        painter.drawRect(option.rect)
        painter.restore()

SubtractionManagerBase, SubtractionManagerForm = loadUI('Subtraction.ui')



class SubtractionManager(SubtractionManagerBase, SubtractionManagerForm):
    def __init__(self, parent=None, guiWindow=None):
        super(SubtractionManager, self).__init__(parent,QtCore.Qt.WindowSystemMenuHint | QtCore.Qt.WindowTitleHint | QtCore.Qt.WindowCloseButtonHint | \
            QtCore.Qt.WindowMinMaxButtonsHint )
        self.setupUi(self)
        self.guiWindow = guiWindow
        self.setWindowIcon(QtGui.QIcon(self.guiWindow.AppContext.get_resource('Icons/Own/subtract.png')))
        self.initSubtractionManager()
        #self.guiWindow.subtractionHelp()

    def accept(self):
        if not self.Subtraction_generateSubtraction_button.isDisabled():
            self.Subtraction_generateSubtraction_button.clicked.emit()
            super(SubtractionManager,self).accept()
        
    

    def initSubtractionManager(self):

        self.Subtraction_dataSet_foreground_comboBox.setModel(self.guiWindow.DataSetModel)
        self.Subtraction_dataSet_background_comboBox.setModel(self.guiWindow.DataSetModel)

       
        self.Subtraction_dataSet_foreground_comboBox.oldIdx = -1
        self.Subtraction_dataSet_background_comboBox.oldIdx = -1

        if self.guiWindow.DataSetModel.rowCount(None)>1:
            self.Subtraction_dataSet_foreground_comboBox.setCurrentIndex(0)
            self.Subtraction_dataSet_foreground_comboBox.oldIdx = 0
            self.Subtraction_dataSet_background_comboBox.setCurrentIndex(1)
            self.Subtraction_dataSet_background_comboBox.oldIdx = 1

        self.Subtraction_dataSet_foreground_comboBox.currentIndexChanged.connect(lambda: self.comboboxChanged(foreground=True))
        self.Subtraction_dataSet_background_comboBox.currentIndexChanged.connect(lambda: self.comboboxChanged(foreground=False))
        
        
        self.DataFileModel_foreground = DataFileModel(self.Subtraction_filenames_foreground_listView,\
            dataSetModel=self.guiWindow.DataSetModel,DataSet_DataSets_listView=self.Subtraction_dataSet_foreground_comboBox,\
                guiWindow = self.guiWindow)

        # Set up model for foreground df view
        self.Subtraction_filenames_foreground_listView.setModel(self.DataFileModel_foreground)
        self.Subtraction_filenames_foreground_listView.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)
        self.Subtraction_filenames_foreground_listView.setDragDropOverwriteMode(False)

        dataSetDelegate = DataSetDelegate() # Painter for DataSet comparison
        dataFileDelegate = DataFileDelegate() # Painter for properties of data files

        self.Subtraction_filenames_foreground_listView.setItemDelegate(dataSetDelegate)
        self.Subtraction_filenames_background_listView.setItemDelegate(dataSetDelegate)

        self.Subtraction_fileAttributs_foreground_listView.setItemDelegate(dataFileDelegate)
        self.Subtraction_fileAttributs_background_listView.setItemDelegate(dataFileDelegate)
        

        # add selection model for smooth selection when drag/dropping
        self.DataFile_foreground_SelectionModel = SelectionModel(self.DataFileModel_foreground)
        
        
        self.Subtraction_filenames_foreground_listView.setSelectionModel(self.DataFile_foreground_SelectionModel)
        
        

        # add file information
        self.DataFile_foreground_infoModel = DataFileInfoModel(DataSet_filenames_listView=self.Subtraction_filenames_foreground_listView,dataSetModel=self.guiWindow.DataSetModel,
        DataSet_DataSets_listView=self.Subtraction_dataSet_foreground_comboBox,dataFileModel=self.DataFileModel_foreground,guiWindow=self.guiWindow)
        self.DataFile_foreground_infoModel.infos = [x for x in subtractionSettings.keys()]
        
        self.Subtraction_fileAttributs_foreground_listView.setModel(self.DataFile_foreground_infoModel)

        self.DataFileModel_foreground.dragDropFinished.connect(self.DataFile_foreground_SelectionModel.onModelItemsReordered)
        self.DataFile_foreground_SelectionModel.selectionChanged.connect(self.updateDataFileLabels)
        


        ##### Repeat above for background
        self.DataFileModel_background = DataFileModel(self.Subtraction_filenames_background_listView,\
            dataSetModel=self.guiWindow.DataSetModel,DataSet_DataSets_listView=self.Subtraction_dataSet_background_comboBox,\
                guiWindow = self.guiWindow)

        # Set up model for background df view
        self.Subtraction_filenames_background_listView.setModel(self.DataFileModel_background)
        self.Subtraction_filenames_background_listView.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)
        self.Subtraction_filenames_background_listView.setDragDropOverwriteMode(False)

        # add selection model for smooth selection when drag/dropping
        self.DataFile_background_SelectionModel = SelectionModel(self.DataFileModel_background)#self.ui.DataSet_DataSets_listView.selectionModel()
        self.DataFileModel_background.dragDropFinished.connect(self.DataFile_background_SelectionModel.onModelItemsReordered)
        self.DataFile_background_SelectionModel.selectionChanged.connect(self.updateDataFileLabels)
        self.Subtraction_filenames_background_listView.setSelectionModel(self.DataFile_background_SelectionModel)
        self.DataFileModel_foreground.dragDropFinished.connect(self.updateDataFileLabels)

        

        # add file information
        self.DataFile_background_infoModel = DataFileInfoModel(DataSet_filenames_listView=self.Subtraction_filenames_background_listView,dataSetModel=self.guiWindow.DataSetModel,
        DataSet_DataSets_listView=self.Subtraction_dataSet_background_comboBox,dataFileModel=self.DataFileModel_background,guiWindow=self.guiWindow)
        self.DataFile_background_infoModel.infos = [x for x in subtractionSettings.keys()]
        self.Subtraction_fileAttributs_background_listView.setModel(self.DataFile_background_infoModel)

        # Connect generation button with method
        self.Subtraction_generateSubtraction_button.clicked.connect(self.generateSubtracted)

        # Check DataSetName on finish editing
        self.Subtraction_dataSet_name_lineEdit.editingFinished.connect(self.checkDataSetName)
        # Call once to update naming from default
        self.checkDataSetName()
        self.checkDataSetCombability()

    def comboboxChanged(self,foreground=True):
        if foreground:
            self.DataFileModel_foreground.updateCurrentDataSetIndex()
        else:
            self.DataFileModel_background.updateCurrentDataSetIndex()
        self.updateDataFileLabels()


    def updateDataFileLabels(self):
        self.DataFile_foreground_infoModel.layoutChanged.emit()
        self.DataFile_background_infoModel.layoutChanged.emit()
        self.checkDataSetCombability()

    def generateSubtracted(self):
        foreground_ds = self.DataFileModel_foreground.getCurrentDataSet()
        background_ds = self.DataFileModel_background.getCurrentDataSet()
        
        temp = foreground_ds-background_ds
        subtracted_ds = GuiDataSet(dataSet=temp,name=self.Subtraction_dataSet_name_lineEdit.text())

        if background_ds[0].type == 'nxs': # If first file is converted
            subtracted_ds.background = [df.original_file.fileLocation for df in background_ds]
            subtracted_ds.convertBeforeSubtract = True
        else:    
            subtracted_ds.background = [df.fileLocation for df in background_ds]
            subtracted_ds.convertBeforeSubtract = False
        
        self.guiWindow.DataSetModel.append(subtracted_ds)
        self.close()
        

    def checkDataSetName(self):
        """If no name is given, simply return "subtracted" """
        if self.Subtraction_dataSet_name_lineEdit.text().strip() == '':
            self.Subtraction_dataSet_name_lineEdit.setText('Subtracted')

    def checkDataSetCombability(self,*args,**kwargs):
        if self.DataFileModel_foreground.rowCount(None)==0 or self.DataFileModel_foreground.rowCount(None)==0:
            self.Subtraction_generateSubtraction_button.setDisabled(True)
            self.DataFileModel_foreground.layoutChanged.emit()
            self.DataFileModel_background.layoutChanged.emit()
            self.guiWindow.DataFileModel.layoutChanged.emit()
            return

        foregroundFiles = self.DataFileModel_foreground.getData()
        backgroundFiles = self.DataFileModel_background.getData()

        maxlength = np.max([len(foregroundFiles),len(backgroundFiles)])
        checks = [] # Check of combability 
        for item in foregroundFiles+backgroundFiles:
            item.BGColor = InvalidColour

        for fore,back in zip(foregroundFiles,backgroundFiles):
            check = subtractable(fore,back)
            totalCheck = np.all([v for v in check.values()])
            checks.append(totalCheck)
            
            if totalCheck:
                fore.BGColor = ValidColour
                back.BGColor = ValidColour
            else:
                fore.BGColor = InvalidColour
                back.BGColor = InvalidColour

            for key,value in check.items():
                setattr(fore,key+'_check',value)
                setattr(back,key+'_check',value)

        # Check if selected datasets are converted
        subtracted = np.any([not X.dataSetModel.item(X.getCurrentDatasetIndex()).background is None for X in [self.DataFileModel_foreground,self.DataFileModel_background]])
        
        if np.all(checks) and len(checks)==maxlength and not subtracted: # subraction is allowed
            self.Subtraction_generateSubtraction_button.setDisabled(False)
        else:
            self.Subtraction_generateSubtraction_button.setDisabled(True)

        self.DataFileModel_foreground.layoutChanged.emit()
        self.DataFileModel_background.layoutChanged.emit()
        self.guiWindow.DataFileModel.layoutChanged.emit()

